"""Download functions."""

import os.path
import typing as ty
from pathlib import Path
from urllib import request

from loguru import logger
from tqdm import tqdm

from koyo.timer import MeasureTimer
from koyo.typing import PathLike


def download_progress(pbar):
    """Wraps tqdm instance.

    Don't forget to close() or __exit__()
    the tqdm instance once you're done with it (easiest using `with` syntax).

    Example
    -------
    >>> with tqdm(...) as p_bar:
    ...     reporthook = download_progress(p_bar)
    ...     request.urlretrieve(reporthook=reporthook)
    """
    last_b = [0]

    def update_to(b=1, bsize=1, tot_size=None):
        """Update progress.

        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tot_size  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tot_size is not None:
            pbar.total = tot_size
        pbar.update((b - last_b[0]) * bsize)
        last_b[0] = b

    return update_to


def download_file(
    link: str,
    filename: ty.Optional[PathLike] = None,
    output_dir: ty.Optional[PathLike] = None,
    unzip: bool = True,
    remove_archive: bool = False,
):
    """Download a file.

    By default, this function download a file to ~/explorer_data.

    Parameters
    ----------
    link : str
        Name of the file to download or url.
    filename : PathLike
        Name of the file to be saved in case of url.
    output_dir : PathLike
        Download file to the path specified.
    unzip : bool | False
        Unzip archive if needed.
    remove_archive : bool | False
        Remove archive after unzip.

    Returns
    -------
    path_to_file : Path
        Path to the downloaded file.
    """
    out_path = Path(__file__).parent
    if "http" in link:
        if filename is None:
            filename = os.path.split(link)[1]
            if "?" in filename:
                filename = filename.split("?")[0]
        assert isinstance(filename, str)
        url = link
    else:
        raise ValueError("Expected a link!")

    output_dir = out_path if not isinstance(output_dir, (str, Path)) else output_dir
    path_to_file = Path(output_dir) / filename
    temp_file = path_to_file.parent / (path_to_file.name + ".temp")
    to_download = not path_to_file.exists()

    # download file if needed
    with MeasureTimer() as timer:
        if to_download:
            logger.info(f"Downloading {path_to_file}")
            # Check if directory exists else creates it
            if not os.path.exists(output_dir):
                logger.info(f"Folder {output_dir} created")
                os.makedirs(output_dir)
            # Download file
            with tqdm(
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
                desc=f"Downloading {filename}...",
                mininterval=1.0,
            ) as p_bar:
                reporthook = download_progress(p_bar)
                _, _ = request.urlretrieve(url, temp_file, reporthook=reporthook)
            temp_file.rename(path_to_file)
        else:
            logger.info(f"File already downloaded ({path_to_file}). Skipping")
    if unzip:
        from koyo.compression import unzip_directory

        path_to_file = unzip_directory(path_to_file, output_dir, remove_archive)

    logger.info(f"Downloaded '{path_to_file}' in {timer()}.")
    return path_to_file



"""zarr.py."""

import numpy as np

from koyo.utilities import is_installed

HAS_ZARR = is_installed("zarr")

if HAS_ZARR:
    import zarr
    from numcodecs import Blosc
    from zarr.storage import ZipStore
else:
    zarr = Blosc = ZipStore = None


def get_chunk_shape_along_axis(array: np.ndarray, axis: int = 0) -> tuple[int, int]:
    """Get chunk size."""
    if array.ndim == 1:
        return (array.shape[0],)
    elif array.ndim == 2:
        if axis == 0:
            return array.shape[0], 1
        elif axis == 1:
            return 1, array.shape[1]
    return 256, 256


def save_array_to_zip(array: np.ndarray, zip_path: str, chunk_size: tuple[int, int] = (256, 256)) -> None:
    """
    Save a 2D NumPy array to a .zip file using zarr + Blosc (zstd) compression.

    Parameters
    ----------
    - array: 2D NumPy array to store
    - zip_path: Path to output .zip file
    - chunk_size: Chunk shape for compression
    """
    if not HAS_ZARR:
        raise ImportError("zarr is not installed. Please install it to use this function.")
    if array.ndim != 2:
        raise ValueError("Only 2D arrays supported")

    # compressor = Blosc(cname="zstd", clevel=5, shuffle=Blosc.BITSHUFFLE)
    compressor = Blosc(cname="lz4", clevel=5, shuffle=Blosc.BITSHUFFLE)
    store = ZipStore(zip_path, mode="w")
    zarr.save_array(store, array, chunks=chunk_size, compressor=compressor)
    store.close()


def load_array_from_zip(zip_path: str) -> np.ndarray:
    """
    Load a 2D NumPy array from a compressed .zip file stored with Zarr.

    Parameters
    ----------
    - zip_path: Path to the .zip file

    Returns
    -------
    - The decompressed NumPy array
    """
    if not HAS_ZARR:
        raise ImportError("zarr is not installed. Please install it to use this function.")

    store = ZipStore(zip_path, mode="r")
    array = zarr.open(store, mode="r")
    result = array[:]
    store.close()
    return result



import os
import sys
from contextlib import contextmanager, suppress

from loguru import logger

DEFAULT_HOOK = None


def debugger_hook(type, value, tb) -> None:
    """Drop into python debugger on uncaught exception."""
    if hasattr(sys, "ps1") or not sys.stderr.isatty():
        # we are in interactive mode, or we don't have a tty-like
        # device, so we call the default hook
        sys.__excepthook__(type, value, tb)
    else:
        import pdb
        import traceback

        # we are NOT in interactive mode, print the exception...
        with suppress(Exception):
            traceback.print_exception(type, value, tb)
            pdb.post_mortem(tb)


def install_debugger_hook() -> None:
    """Activate the debugger hook."""
    global DEFAULT_HOOK
    os.environ["KOYO_DEV_MODE"] = "1"
    os.environ["DEV_MODE"] = "1"

    if DEFAULT_HOOK is None:
        DEFAULT_HOOK = sys.excepthook
    sys.excepthook = debugger_hook


def uninstall_debugger_hook() -> None:
    """Deactivate the debugger hook."""
    global DEFAULT_HOOK
    os.environ["KOYO_DEV_MODE"] = "0"
    os.environ["DEV_MODE"] = "0"

    if DEFAULT_HOOK is not None:
        sys.excepthook = DEFAULT_HOOK
        DEFAULT_HOOK = None


def logger_hook(type, value, tb) -> None:
    """Logger hook."""
    import traceback

    # we are NOT in interactive mode, print the exception...
    with suppress(Exception):
        traceback.print_exception(type, value, tb)
        logger.exception(value)


def install_logger_hook() -> None:
    """Activate the debugger hook."""
    global DEFAULT_HOOK
    os.environ["KOYO_LOG_MODE"] = "1"
    os.environ["LOG_MODE"] = "1"

    if DEFAULT_HOOK is None:
        DEFAULT_HOOK = sys.excepthook
    sys.excepthook = logger_hook


def uninstall_logger_hook() -> None:
    """Deactivate the debugger hook."""
    global DEFAULT_HOOK
    os.environ["KOYO_LOG_MODE"] = "0"
    os.environ["LOG_MODE"] = "0"

    if DEFAULT_HOOK is not None:
        sys.excepthook = DEFAULT_HOOK
        DEFAULT_HOOK = None


@contextmanager
def catch_if_debug():
    """Catch exception if debugging."""
    if os.environ.get("DEV_MODE", "0") == "0":
        try:
            yield
        except Exception:
            pass
    else:
        yield



"""Compression utilities."""

import os
import zipfile

from loguru import logger
from tqdm import tqdm

from koyo.typing import PathLike

DECOMPRESSION_FORMATS = ["gz", ".gz", "*.gz"]


def unzip_directory(path_to_zip: PathLike, output_dir: PathLike, remove_archive: bool = True) -> PathLike:
    """Unzip directory."""
    logger.debug("Unzipping directory...")
    # unzip archive
    zip_file_object = zipfile.ZipFile(path_to_zip, "r")
    zip_file_object.extractall(path=output_dir)
    path_to_file = os.path.join(output_dir, zip_file_object.namelist()[0])
    zip_file_object.close()
    logger.debug("Unzipped directory")

    if remove_archive:  # Remove archive :
        logger.debug(f"Archive {path_to_zip} removed")
        os.remove(path_to_zip)
    return path_to_file


def zip_directory(path_to_directory: PathLike, output_dir: PathLike) -> PathLike:
    """Zip directory."""
    logger.debug("Zipping directory...")
    # zip directory
    zip_file_object = zipfile.ZipFile(output_dir, "w", zipfile.ZIP_DEFLATED)
    files = list(os.walk(path_to_directory))
    for root, _, files in tqdm(files, desc="Zipping", unit="file"):
        for file in files:
            zip_file_object.write(
                os.path.join(root, file), os.path.relpath(os.path.join(root, file), path_to_directory)
            )
    zip_file_object.close()
    logger.debug("Zipped directory")
    return output_dir



"""Check for latest release."""

from __future__ import annotations

import typing as ty

import requests
from loguru import logger

from koyo.timer import format_datetime_ago


class DownloadDict(ty.TypedDict):
    """Download dict."""

    filename: str
    version: str
    download_url: str


LatestVersion = dict[str, DownloadDict]


def get_target() -> str | None:
    """Get target based on the platform."""
    import platform

    if platform.system() == "Windows":
        return "win_amd64"
    elif platform.system() == "Darwin":
        if platform.processor() == "arm":
            return "macosx_arm64"
        return "macosx_x86_64"
    return None


def get_latest_release(user: str = "vandeplaslab", package: str = "koyo") -> str:
    """Get latest release from GitHub."""
    data = get_latest_git(user, package)
    if "tag_name" in data:
        return data["tag_name"]
    return ""


def get_latest_git(user: str = "vandeplaslab", package: str = "koyo") -> dict[str, ty.Any]:
    """Get latest release from GitHub."""
    response = requests.get(f"https://api.github.com/repos/{user}/{package}/releases/latest")
    if response.status_code == 200:
        data = response.json()
        return data
    return {}


def format_version(data: dict) -> str:
    """Format version."""
    git_version = data.get("tag_name", "")
    if not git_version:
        return "No changelog information available."
    published_at = format_datetime_ago(data.get("published_at", ""))
    name = data.get("name", "")
    body = data.get("body", "")
    return f"# {name}\n**Published:** {published_at}\n**Change log**\n\n{body}"


def is_new_version_available(
    current_version: str, user: str = "vandeplaslab", package: str = "koyo", data: dict[str, ty.Any] | None = None
) -> tuple[bool, str]:
    """Check whether there is a new version available."""
    import requests.exceptions
    from packaging import version

    # get latest version from GitHub
    if data is None:
        data = {}
        try:
            data = get_latest_git(user, package)
        except requests.exceptions.ConnectionError:
            logger.trace("There was a connection error - could not retrieve information.")
        except requests.exceptions.RequestException:
            logger.trace("There was some problem in retrieving information.")
        except Exception:
            logger.trace("Another exception occur - not sure how to handle this!")

    git_version = data.get("tag_name", "")
    if not git_version:
        return False, "Could not retrieve latest version."
    new_available = version.parse(git_version) > version.parse(current_version)
    if new_available:
        published_at = format_datetime_ago(data.get("published_at", ""))
        return True, (
            f"New version available: <b>{git_version}</b> made available <b>{published_at}</b>."
            f" You are using version <b>{current_version}</b>"
        )
    return False, f"You are using the latest version: '{current_version}'."



"""Definitions of types."""

import sys
import typing as ty
from pathlib import Path

if ty.TYPE_CHECKING:
    import numpy as np
    from scipy import sparse


__all__ = ("ArrayLike", "CompressedSparseArray", "PathLike", "SparseArray")

ArrayLike = ty.Union[ty.List, ty.Tuple, "np.ndarray", "sparse.csc_matrix", "sparse.csr_matrix"]
SimpleArrayLike = ty.TypeVar("SimpleArrayLike", ty.List, "np.ndarray", ty.Iterable)
CompressedSparseArray = ty.Union["sparse.csc_matrix", "sparse.csr_matrix"]
SparseArray = ty.Union["sparse.csc_matrix", "sparse.csr_matrix", "sparse.coo_matrix"]
PathLike = ty.Union[str, Path]


if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    # in 3.11+, using the below class in an f-string would put the enum name instead of its value
    from enum import Enum

    class StrEnum(str, Enum):
        pass



"""Decorators."""

from __future__ import annotations

import functools
import logging
import time
import typing as ty
import warnings
from functools import partial, wraps


def retry(
    func: ty.Callable | None = None,
    exception: Exception = Exception,
    n_tries: int = 5,
    delay: float = 5,
    backoff: int = 1,
    logger: bool = False,
) -> ty.Callable:
    """Retry decorator with exponential backoff.

    Parameters
    ----------
    func : typing.Callable, optional
        Callable on which the decorator is applied, by default None
    exception : Exception or tuple of Exceptions, optional
        Exception(s) that invoke retry, by default Exception
    n_tries : int, optional
        Number of tries before giving up, by default 5
    delay : int, optional
        Initial delay between retries in seconds, by default 5
    backoff : int, optional
        Backoff multiplier e.g. value of 2 will double the delay, by default 1
    logger : bool, optional
        Option to log or print, by default False

    Returns
    -------
    typing.Callable
        Decorated callable that calls itself when exception(s) occur.

    Notes
    -----
    Taken from https://stackoverflow.com/a/61093779/4364202

    Examples
    --------
    >>> import random
    >>> @retry(exception=Exception, n_tries=4)
    ... def test_random(text):
    ...    x = random.random()
    ...    if x < 0.5:
    ...        raise Exception("Fail")
    ...    else:
    ...        print("Success: ", text)
    >>> test_random("It works!")
    """
    if func is None:
        return partial(
            retry,
            exception=exception,
            n_tries=n_tries,
            delay=delay,
            backoff=backoff,
            logger=logger,
        )

    @wraps(func)
    def wrapper(*args, **kwargs):
        ntries, ndelay = n_tries, delay

        while ntries > 1:
            try:
                return func(*args, **kwargs)
            except exception as e:
                msg = f"{e!s}, Retrying in {ndelay} seconds..."
                if logger:
                    logging.warning(msg)
                else:
                    print(msg)
                time.sleep(ndelay)
                ntries -= 1
                ndelay *= backoff

        return func(*args, **kwargs)

    return wrapper


def deprecated(func: ty.Callable, context: str = "") -> ty.Callable:
    """A decorator which can be used to mark functions as deprecated.

    It will result in a warning being emitted when the function is used.
    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter("always", DeprecationWarning)  # turn off filter
        warnings.warn(
            f"Call to deprecated function {func.__name__}. {context}",
            category=DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter("default", DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    return new_func


class classproperty:
    """
    Decorator that converts a method with a single cls argument into a property
    that can be accessed directly from the class.
    """

    def __init__(self, method=None):
        self.fget = method

    def __get__(self, instance, cls=None):
        return self.fget(cls)

    def getter(self, method):
        self.fget = method
        return self



"""Random number utilities."""

from __future__ import annotations

from contextlib import contextmanager

import numpy as np


def get_random_seed() -> int:
    """Get random seed."""
    return np.random.randint(0, np.iinfo(np.int32).max - 1, 1)[0]


def get_random_state(n: int = 1) -> int | list[int]:
    """Retrieve random state(s)."""
    from random import randint

    state = [randint(0, 2**32 - 1) for _ in range(n)]
    return state if n > 1 else state[0]


@contextmanager
def temporary_seed(seed: int, skip_if_negative_one: bool = False):
    """Temporarily set numpy seed."""
    if skip_if_negative_one and seed == -1:
        yield
        return
    state = np.random.get_state()
    np.random.seed(seed)
    try:
        yield
    finally:
        np.random.set_state(state)



"""Visuals."""

from __future__ import annotations

import math
import typing as ty
import warnings
from contextlib import contextmanager, suppress

import matplotlib.pyplot as plt
import numba as nb
import numpy as np

from koyo.image import clip_hotspots
from koyo.typing import PathLike
from koyo.utilities import is_above_version

if ty.TYPE_CHECKING:
    import pandas as pd
    from matplotlib.collections import LineCollection
    from matplotlib.colorbar import Colorbar
    from matplotlib.image import AxesImage
    from PIL import Image
    from seaborn.matrix import _HeatMapper

MATPLOTLIB_ABOVE_3_5_2 = is_above_version("matplotlib", "3.5.2")


def tight_layout(fig: plt.Figure) -> None:
    """Apply tight layout to the figure."""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        fig.tight_layout()


def close_mpl_figure(fig: plt.Figure) -> None:
    """Close figure without raising exception."""
    with suppress(Exception):
        plt.close(fig)


def close_pil_image(fig: Image) -> None:
    """Close PIL image without raising exceptions."""
    with suppress(Exception):
        fig.close()
        del fig


@contextmanager
def disable_bomb_protection() -> None:
    """Disable PIL decompression bomb protection."""
    from PIL import Image

    max_size = Image.MAX_IMAGE_PIXELS
    # disable decompression bomb protection
    Image.MAX_IMAGE_PIXELS = 30_000 * 30_000  # 30k x 30k pixels
    yield
    # re-enable decompression bomb protection
    Image.MAX_IMAGE_PIXELS = max_size


def save_rgb(path: PathLike, image: np.ndarray) -> None:
    """Save image a RGB using PIL."""
    from PIL import Image

    Image.fromarray(image).save(path)


def save_gray(path: PathLike, image: np.ndarray, multiplier: int = 255) -> None:
    """Save image as grayscale using PIL."""
    from PIL import Image

    Image.fromarray(image * multiplier).convert("L").save(path)


def set_tick_fmt(ax: plt.Axes, use_offset: bool = False, axis: str = "both") -> plt.Axes:
    """Set tick format to control whether scientific notation is shown."""
    ax.ticklabel_format(axis=axis, style="scientific" if use_offset else "plain", useOffset=use_offset)
    return ax


def despine(ax: plt.Axes, orientation: str) -> plt.Axes:
    """Remove spines from 1D plots."""
    plt.setp(ax.xaxis.get_majorticklines(), visible=False)
    plt.setp(ax.xaxis.get_minorticklines(), visible=False)
    plt.setp(ax.yaxis.get_majorticklines(), visible=False)
    plt.setp(ax.yaxis.get_minorticklines(), visible=False)
    plt.setp(ax.get_xticklabels(), visible=False)
    plt.setp(ax.get_yticklabels(), visible=False)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    if orientation in ["horizontal", "horz", "h"]:
        ax.spines["left"].set_visible(False)
    else:
        ax.spines["bottom"].set_visible(False)
    return ax


def fig_to_pil(fig: plt.Figure) -> Image:
    """Convert a Matplotlib figure to a PIL Image and return it."""
    import io

    import matplotlib.pyplot as plt
    from PIL import Image

    buf = io.BytesIO()
    fig.savefig(buf, bbox_inches="tight")
    plt.close(fig)
    buf.seek(0)
    img = Image.open(buf)
    return img


def show_fig(fig: plt.Figure) -> None:
    """Show figure."""
    image = fig_to_pil(fig)
    image.show()


def pil_to_fig(image: Image) -> plt.Figure:
    """Convert a PIL Image to a Matplotlib figure and return it."""
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=(image.width / 72, image.height / 72))
    ax = fig.add_axes([0, 0, 1, 1])
    ax.set_axis_off()
    ax.imshow(image)
    return fig


def clear_axes(i: int, axs):
    """Clear axes that were not used."""
    if i < len(axs) - 1:
        for _i in range(i + 1, len(axs)):
            axs[_i].axis("off")


def auto_clear_axes(axs):
    """Automatically clear axes. This checks whether there are any lines or collections."""
    if isinstance(axs, np.ndarray):
        axs = axs.flatten()
    for ax in axs:
        if not ax.lines and not ax.collections:
            ax.axis("off")


@nb.njit()
def convert_to_vertical_line_input(x: np.ndarray, y: np.ndarray):
    """Convert two arrays of the same length to list of [x, 0, y, y] vertices used by the matplotlib LineCollection."""
    return [[(x[i], 0), (x[i], y[i])] for i in range(len(x))]


def vertices_to_collection(
    x: np.ndarray,
    y: np.ndarray,
    color: str = "b",
    line_width=3,
    line_style="-",
    alpha: float = 1.0,
) -> LineCollection:
    """Convert list of [x, 0, y, y] vertices to line collection consumed by matplotlib."""
    from matplotlib.collections import LineCollection

    xy_values = convert_to_vertical_line_input(x, y)
    line_col = LineCollection(
        xy_values,
        colors=[color] * len(x),
        linewidths=line_width,
        linestyles=line_style,
        alpha=alpha,
    )
    return line_col


def plot_centroids(
    x: np.ndarray, y: np.ndarray, x_label: str = "", y_label: str = "", title: str = "", xmin: float | None = None, xmax: float | None = None
):
    """Plot NMF/PCA component contribution."""
    color = plt.rcParams["text.color"]

    fig, ax = plt.subplots(figsize=(10, 6))
    lc = vertices_to_collection(x, y, color=color, line_width=1.5)
    ax.add_collection(lc)
    ax.set_ylabel(y_label, fontsize=14)
    ax.tick_params(axis="both", which="major", labelsize=12)
    ax.set_ylim(np.min(y) * 1.05, np.max(y) * 1.05)  # needed because centroids are not properly setting min/max
    if xmin is not None and xmax is not None:
        ax.set_xlim(xmin, xmax)
    ax.set_xlabel(x_label, fontsize=14)
    ax.set_title(title, fontsize=16)
    return fig, ax


def compute_divider(value: float) -> int:
    """Compute divider."""
    divider = 1_000_000_000
    value = abs(value)
    if value == 0:
        return 1
    with suppress(ZeroDivisionError), warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        while value == value % divider:
            divider = divider / 1_000
    return len(str(int(divider))) - len(str(int(divider)).rstrip("0"))


def convert_divider_to_str(value: float, exp_value: int) -> str:
    """Convert divider to string."""
    value = float(value)
    if exp_value in [0, 1, 2]:
        if abs(value) < 0.0001:
            return f"{value:.6G}"
        if abs(value) < 0.01:
            return f"{value:.4G}"
        if abs(value) <= 1:
            return f"{value:.2G}"
        elif abs(value) <= 1000:
            if value.is_integer():
                return f"{value:.0F}"
            return f"{value:.1F}"
    elif exp_value in [3, 4, 5]:
        return f"{value / 1000:.1f}k"
    elif exp_value in [6, 7, 8]:
        return f"{value / 1000000:.1f}M"
    elif exp_value in [9, 10, 11, 12]:
        return f"{value / 1000000000:.1f}B"


def y_tick_fmt(x, pos=None) -> str:
    """Y-tick formatter."""
    return convert_divider_to_str(x, compute_divider(x))


def get_intensity_formatter():
    """Simple intensity formatter."""
    from matplotlib.ticker import FuncFormatter

    return FuncFormatter(y_tick_fmt)


def set_intensity_formatter(ax):
    """Set intensity formatter on axes."""
    ax.yaxis.set_major_formatter(get_intensity_formatter())


def add_contours(
    ax: plt.Axes, contours: np.ndarray | dict[str, np.ndarray], line_width: float = 1.0, color: str | None = None
) -> None:
    """Add contours to the axes."""
    if contours is None:
        return
    if isinstance(contours, np.ndarray):
        contours = {"": contours}
    color = color if color is not None else plt.rcParams["text.color"]
    for _key, contour in contours.items():
        ax.plot(contour[:, 0], contour[:, 1], lw=line_width, color=color, zorder=100)


def _sort_contour_order(contours: dict[str, np.ndarray]) -> tuple[dict[str, tuple[int, int]], dict[str, np.ndarray]]:
    # get min x, y for each contour
    order = {}
    for key, contour in contours.items():
        order[key] = (np.min(contour[:, 1]), np.min(contour[:, 0]))
    contours = dict(sorted(contours.items(), key=lambda x: order[x[0]]))
    return order, contours


def _get_alternate_locations(contours: dict[str, np.ndarray]) -> dict[str, str]:
    """Get alternative locations."""
    _, contours = _sort_contour_order(contours)
    locations = {}
    is_top = True
    for key in contours:
        locations[key] = "top" if is_top else "bottom"
        is_top = not is_top
    return locations


def add_contour_labels(
    ax: plt.Axes,
    contours: np.ndarray | dict[str, np.ndarray],
    labels: str | dict[str, str],
    font_size: int = 12,
    color: str | None = None,
    where: ty.Literal["top", "bottom", "alternate"] | tuple[int, int] = "alternate",
    locations: dict[str, str] | None = None,
) -> None:
    """Add labels to the contours."""
    if contours is None:
        return

    y_offset = font_size
    is_top = where == "top"
    is_alt = where == "alternate"
    if is_alt:
        is_top = True

    locations = locations or {}
    if isinstance(contours, np.ndarray):
        contours = {"": contours}
    if isinstance(labels, str):
        labels = {"": labels}
    if isinstance(locations, str):
        locations = {"": locations}

    # get min x, y for each contour
    if where == "alternate":
        _, contours = _sort_contour_order(contours)

    color = color if color is not None else plt.rcParams["text.color"]
    new_y_ax = 0
    for key, contour in contours.items():
        # find horizontal center of the contour
        xs = contour[:, 0]
        xmin, xmax = xs.min(), xs.max()
        x = xmin + (xmax - xmin) / 2
        location = locations.get(key, "top" if is_top else "bottom")
        is_top = location == "top"

        # find vertical top of the contour
        y_min = np.min(contour[:, 1])
        y_max = np.max(contour[:, 1])
        y = y_min if is_top else y_max + y_offset
        if not isinstance(location, str):
            if len(location) == 2:
                x_, y_ = location
            elif len(location) == 3:
                where, x_offset_, y_offset_ = location
                y_ = y_min if where == "top" else y_max + y_offset
                x_ = x + x_offset_
                y_ += y_offset_
        else:
            x_, y_ = x, y
        ax.text(
            x_,
            y_,
            labels[key],
            fontsize=font_size,
            color=color,
            va="bottom" if is_top else "top",
            ha="center",
            zorder=100,
        )
        new_y_ax = min(y, new_y_ax)  # if is_top else min(y, new_y_ax)
        if is_alt:
            is_top = not is_top
    # set y limit to the maximum y value
    ax_y_lim = ax.get_ylim()
    y_min, y_max = np.min(ax_y_lim), np.max(ax_y_lim)
    y_min = min(y_min, new_y_ax)
    padding = y_offset * 8
    if where == "top":
        ax_y_lim = (y_max + padding / 2, y_min - padding)
    elif where == "bottom":
        ax_y_lim = (y_max + padding, y_min - padding / 2)
    else:
        ax_y_lim = (y_max + padding, y_min - padding)
    ax.set_ylim(ax_y_lim)


def add_ax_colorbar(mappable):
    """Add colorbar to axis."""
    import matplotlib.pyplot as plt
    from mpl_toolkits.axes_grid1 import make_axes_locatable

    last_axes = plt.gca()
    ax = mappable.axes
    fig = ax.figure
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = fig.colorbar(mappable, cax=cax)
    plt.sca(last_axes)
    return cbar


def make_legend_handles(
    names: list[str],
    colors: list[str],
    kind: list[str] = "line",
    width: list[float] = 3,
    extra_kws: dict | None = None,
):
    """Create custom legend."""
    from matplotlib.lines import Line2D
    from matplotlib.patches import Patch

    extra_kws = extra_kws or {}
    if isinstance(colors, (str, np.ndarray)):
        colors = [colors] * len(names)
    if isinstance(kind, str):
        kind = [kind] * len(names)
    if isinstance(width, (float, int)):
        width = [width] * len(names)
    handles = []
    for name, color, kind_, width_ in zip(names, colors, kind, width):
        kws = extra_kws.get(name, {})
        if kind_ == "line":
            handles.append(Line2D([0], [0], color=color, lw=width_, label=name, **kws))
        elif kind_ == "patch":
            handles.append(Patch(facecolor=color, edgecolor=color, label=name, **kws))
        elif kind_ == "patch":
            handles.append(Patch(facecolor=color, edgecolor=color, label=name, **kws))
        else:
            raise ValueError(f"Unknown kind {kind_}")
    return handles


def add_patches(axs: list, windows: list[tuple[float, float]], colors: list | None = None, alpha: float = 0.5):
    """Add rectangular patches associated with the peak."""
    from matplotlib.patches import Rectangle

    if colors is None:
        colors = plt.cm.get_cmap("viridis", len(axs)).colors
    assert len(axs) == len(windows) == len(colors), "The number of axes does not match the number of windows."

    for ax, (xmin, xmax), color in zip(axs, windows, colors):
        ax.add_patch(Rectangle((xmin, 0), xmax - xmin, ax.get_ylim()[1], alpha=alpha, color=color))


def add_scalebar(ax, px_size: float | None, color="k"):
    """Add scalebar to figure."""
    try:
        from matplotlib_scalebar.scalebar import ScaleBar
    except ImportError:
        print("matplotlib-scalebar not installed. Please install it using 'pip install matplotlib-scalebar'")
        return

    scalebar = ScaleBar(px_size, "um", frameon=False, color=color, font_properties={"size": 20})
    ax.add_artist(scalebar)
    return scalebar


def add_legend(
    fig: plt.Figure,
    ax: plt.Axes,
    legend_palettes: dict[str, str] | dict[str, dict[str, str]],
    fontsize: float = 14,
    labelsize: float = 16,
    x_pad: float = 0.01,
    sort_legend: bool = True,
) -> None:
    """Add legend to the plot.

    Parameters
    ----------
    fig
        Figure to add legend to.
    ax
        Axes to add legend to.
    legend_palettes
        Dictionary of tag name to tag to color mapping. The mapping should reflect all labels and colors that should be
        added to the legend.
    fontsize
        Font size of the legend labels.
    labelsize
        Font size of the legend title.
    x_pad : float
        Padding between the legend and the plot.
    sort_legend : bool
        Whether to sort the legend by tag name.
    """
    from matplotlib.patches import Patch
    from natsort import natsorted, ns

    def _make_legend(n_col=1, loc="best"):
        return n_col, ax.legend(
            handles=lh,
            loc=loc,
            title=tag,
            frameon=False,
            handlelength=1.2,
            handleheight=1.4,
            fontsize=fontsize,
            title_fontsize=labelsize,
            ncol=n_col,
        )

    # check if legend_palettes is a nested dictionary
    if not all(isinstance(v, dict) for v in legend_palettes.values()):
        legend_palettes = {"": legend_palettes}

    n_palettes = len(legend_palettes) > 1
    rend = fig.canvas.get_renderer()
    x_offset = ax.get_tightbbox(rend).transformed(ax.transAxes.inverted()).xmax + x_pad
    x_widest, y_offset = 0, 1
    if ax.get_title():
        y_offset += ax.title.get_window_extent(rend).transformed(ax.transAxes.inverted()).height
    # calculate maximum height
    max_height = 1.0
    xaxis = ax.get_xaxis()
    ticks = xaxis.get_ticklabels()
    if ticks:
        # find tick with longest label
        index = np.argmax([len(tick.get_text()) for tick in xaxis.get_ticklabels()])
        # get tick
        tick = xaxis.get_ticklabels()[index]
        max_height += tick.get_window_extent(rend).transformed(ax.transAxes.inverted()).height

    for tag, tag_to_color in legend_palettes.items():
        tag_to_color_ = natsorted(tag_to_color.keys(), alg=ns.REAL) if sort_legend else list(tag_to_color.keys())
        lh = [Patch(facecolor=tag_to_color[tag], label=tag) for tag in tag_to_color_]
        n_col = 1
        while True:
            n_col, leg = _make_legend(n_col, loc=(x_offset, y_offset))
            bb = leg.get_window_extent(rend).transformed(ax.transAxes.inverted())
            if bb.height <= max_height:
                break
            n_col += 1
        # if bb.height > 1:
        #     n_col, leg = _make_legend(2, (x_offset, y_offset))
        #     bb = leg.get_window_extent(rend).transformed(ax.transAxes.inverted())
        y_offset -= bb.height
        x_widest = max(x_widest, bb.width)
        if y_offset < 0 and n_palettes:
            x_offset += x_widest + 0.02
            y_offset = 1 - bb.height
        _, leg = _make_legend(n_col, (x_offset, y_offset))
        if not MATPLOTLIB_ABOVE_3_5_2:
            ax.add_artist(leg)


def find_text_color(base_color, dark_color="black", light_color="white", coef_choice=0):
    """
    Takes a background color and returns the appropriate light or dark text color.
    Users can specify the dark and light text color, or accept the defaults of 'black' and 'white'
    base_color: The color of the background. This must be
        specified in RGBA with values between 0 and 1 (note, this is the default
        return value format of a call to base_color = cmap(number) to get the
        color corresponding to a desired number). Note, the value of `A` in RGBA
        is not considered in determining light/dark.
    dark_color: Any valid matplotlib color value.
        Function will return this value if the text should be colored dark
    light_color: Any valid matplotlib color value.
        Function will return this value if thet text should be colored light.
    coef_choice: slightly different approaches to calculating brightness. Currently two options in
        a list, user can enter 0 or 1 as list index. 0 is default.
    """
    # Coefficients:
    # option 0: http://www.nbdtech.com/Blog/archive/2008/04/27/Calculating-the-Perceived-Brightness-of-a-Color.aspx
    # option 1: http://stackoverflow.com/questions/596216/formula-to-determine-brightness-of-rgb-color
    coef_options = [
        np.array((0.241, 0.691, 0.068, 0)),
        np.array((0.299, 0.587, 0.114, 0)),
    ]

    coefs = coef_options[coef_choice]
    rgb = np.array(base_color) * 255
    brightness = np.sqrt(np.dot(coefs, rgb**2))

    # Threshold from option 0 link; determined by trial and error.
    # base is light
    if brightness > 130:
        return dark_color
    return light_color


def stitch_mosaic(filename: str, filelist, n_cols: int):
    """Merge images of the same size and shape."""
    import math

    from PIL import Image

    shape = []
    for _filename in filelist:
        with Image.open(_filename) as file:
            shape.append((file.height, file.width))
    shape = np.array(shape)
    height, width = np.max(shape, axis=0)

    n_rows = math.ceil(len(filelist) / n_cols)

    dst = Image.new("RGB", (width * n_cols, height * n_rows))
    k = 0  # image counter
    for i in range(n_rows):
        y = height * i
        for j in range(n_cols):
            try:
                with Image.open(filelist[k]) as im:
                    x = width * j
                    dst.paste(im, (x, y))
                k += 1
            except IndexError:
                break
    dst.save(filename)
    del dst


def make_image_plot(heatmap, outfname: str | None = None, close: bool = False):
    """Generate simple heatmap."""
    from koyo.utilities import calculate_quantile_without_zeros

    fig, ax = plt.subplots(figsize=(12, 12))
    ax.imshow(heatmap, aspect="equal", vmax=calculate_quantile_without_zeros(heatmap, 0.995))
    ax.axis("off")

    if outfname is not None:
        fig.savefig(outfname, bbox_inches="tight", pad_inches=0.1, dpi=150)

    if close:
        plt.close(fig)


def get_row_col(count, n_rows: int) -> tuple[int, int]:
    """Get number of rows and columns."""
    n_cols = math.ceil(count / n_rows)
    return n_rows, n_cols


def add_label(
    ax,
    label: str,
    x: float = 0.9,
    y: float = 0.98,
    label_color="w",
    font_size=14,
    font_weight="normal",
    va: str = "top",
    ha: str = "left",
    bbox: dict[str, ty.Any] | None = None,
    zorder: int = 10,
):
    """Add label to the image."""
    ax.text(
        x,
        y,
        label,
        transform=ax.transAxes,
        fontsize=font_size,
        fontweight=font_weight,
        verticalalignment=va,
        horizontalalignment=ha,
        color=label_color,
        bbox=bbox,
        zorder=zorder,
    )


def get_ticks_with_unit(
    min_val: float, max_val: float, unit: str | None = None, n: int | None = None
) -> tuple[list[float], list[str]]:
    """Get ticks for specified min/max range."""
    if not unit:
        if n is None:
            ticks, tick_labels = [min_val, max_val], [y_tick_fmt(min_val), y_tick_fmt(max_val)]
        else:
            between = np.linspace(min_val, max_val, n)
            ticks = between.tolist()
            tick_labels = [y_tick_fmt(t) for t in between]
    else:
        # if n is even, let's add one more tick
        if n is None:
            n = 3
        if n % 2 == 0:
            n += 1
        between = np.linspace(min_val, max_val, n)
        ticks = between.tolist()
        tick_labels = [y_tick_fmt(t) for t in between]
        # in the middle insert the unit
        middle = n // 2
        ticks[middle] = unit
    return ticks, tick_labels


def update_colorbar(
    cbar: Colorbar,
    max_val: float,
    min_val: float = 0.0,
    ticks: ty.Sequence[float] | None = None,
    ticklabels: ty.Sequence[str] | None = None,
) -> None:
    """Update colorbar."""
    if not cbar:
        return
    cbar.mappable.set_clim(min_val, max_val)
    cbar.set_ticks(ticks)
    if ticklabels:
        cbar.set_ticklabels(ticklabels)


def inset_colorbar(
    ax: plt.Axes,
    im,
    ticks: ty.Sequence[float] | None = None,
    ticklabels: ty.Sequence[str] | None = None,
    xpos: float = 0.03,
    ypos: float = 0.05,
    labelcolor: str = "auto",
    edgecolor: str = "auto",
    width: str = "30%",
    height: str = "4%",
    orientation: str = "horizontal",
    labelsize: int = 16,
    label: str = "",
    **_kwargs: ty.Any,
):
    """Add colorbar to axes."""

    def _parse_perc(value):
        if "%" in value:
            return float(value.replace("%", "")) / 100
        return float(value)

    if labelcolor == "auto":
        labelcolor = plt.rcParams["text.color"]
    if edgecolor == "auto":
        edgecolor = plt.rcParams["text.color"]

    width = _parse_perc(width)
    height = _parse_perc(height)
    cax = ax.inset_axes([xpos, ypos, width, height])
    cax.tick_params(labelcolor=labelcolor, labelsize=labelsize)
    cbar = plt.colorbar(im, cax=cax, orientation=orientation, pad=0.1, ticks=ticks)
    cbar.outline.set_edgecolor(edgecolor)
    cbar.outline.set_linewidth(1)
    if ticklabels:
        cbar.set_ticklabels(ticklabels)
    if label:
        cbar.set_label(label, size=labelsize + 2)
    return ax, cax, cbar


def fix_style(style: str) -> str:
    """Fix style so that it is compatible with matplotlib > v.3.6.0."""
    from matplotlib.style import available

    if style == "seaborn-grid":
        style = "seaborn-ticks"
    if style == "default" and "default" not in available:
        style = "seaborn-ticks"
    if style == "dark":
        style = "dark_background"
    if style.startswith("seaborn"):
        if "v0_8" not in style and "seaborn-v0_8" in available:
            style = style.replace("seaborn", "seaborn-v0_8")
    if style != "default":
        assert style in available, f"Style '{style}' not available. Available styles: {available}"
    return style


def shorten_style(style: str) -> str:
    """Shorten style name."""
    style = style.replace("seaborn-v0_8-", "s-")  # seaborn style is too long
    style = style.replace("seaborn-", "s-")
    style = style.replace("seaborn", "s")
    style = style.replace("dark_background", "dark")
    return style


def _override_seaborn_heatmap_annotations():
    from seaborn.matrix import _HeatMapper

    if not hasattr(_HeatMapper, "_annotate_heatmap_old"):
        _HeatMapper._annotate_heatmap_old = _HeatMapper._annotate_heatmap
    _HeatMapper._annotate_heatmap = _annotate_heatmap


def _annotate_heatmap(g: _HeatMapper, ax: plt.Axes, mesh) -> None:
    from seaborn.utils import relative_luminance

    mesh.update_scalarmappable()
    height, width = g.annot_data.shape
    xpos, ypos = np.meshgrid(np.arange(width) + 0.5, np.arange(height) + 0.5)
    for x, y, m, color, val in zip(
        xpos.flat, ypos.flat, mesh.get_array().flat, mesh.get_facecolors(), g.annot_data.flat
    ):
        if m is not np.ma.masked:
            lum = relative_luminance(color)
            text_color = ".15" if lum > 0.408 else "w"
            annotation = f"{val:.2f}"
            if val > 0:
                annotation = "1" if annotation in "1.00" else annotation.replace("0.", ".")
            else:
                annotation = "-1" if annotation in "-1.00" else annotation.replace("0.", ".")
            text_kwargs = {"color": text_color, "ha": "center", "va": "center"}
            text_kwargs.update(g.annot_kws)
            ax.text(x, y, annotation, **text_kwargs)


def _plot_image(
    array: np.ndarray,
    colormap: str = "viridis",
    colorbar: bool = True,
    clip: bool = False,
    is_normalized: bool = False,
    title: str = "",
    as_title: bool = True,
    quantile: float = 0.995,
    min_val: float | None = None,
    max_val: float | None = None,
    figsize: tuple[float, float] = (6, 6),
    border_color: str | None = None,
    title_color: str | None = None,
) -> tuple[plt.Figure, plt.Axes]:
    if min_val is None:
        min_val = np.min(array[np.isfinite(array)])
    if clip:
        array = clip_hotspots(array, quantile)
    if not max_val and is_normalized:
        max_val = np.quantile(array[np.isfinite(array)], quantile) if quantile else np.max(array[np.isfinite(array)])
    elif not max_val:
        max_val = np.max(array[np.isfinite(array)])

    kws = {}
    if array.dtype in [np.uint, np.uint8, bool]:
        kws["interpolation"] = "nearest"

    text_color = plt.rcParams["text.color"]
    ticks, tick_labels = get_ticks_with_unit(min_val, max_val)
    fig, ax = plt.subplots(figsize=figsize)
    img = ax.imshow(array, vmax=max_val, cmap=colormap, aspect="equal", **kws)
    if colorbar:
        _, _, cbar = inset_colorbar(
            ax,
            img,
            ticks=ticks,
            ticklabels=tick_labels,
            width="76%",
            xpos=0.12,
            ypos=-0.05,
            labelcolor=text_color,
            edgecolor=text_color,
        )
        cbar.mappable.set_clim(min_val, max_val)
    if as_title:
        ax.set_title(title, color=text_color, fontsize=16)
    else:
        add_label(
            ax,
            title,
            x=0.01,
            y=0.95,
            color="k",
            size=16,
            bbox={"boxstyle": "square", "facecolor": "white", "alpha": 0.75, "lw": 0},
        )
    if title_color:
        ax.title.set_color(title_color)
    if border_color:
        fig.patch.set_linewidth(2)
        fig.patch.set_edgecolor(border_color)
    return fig, ax


def _plot_or_update_image(
    ax: plt.Axes,
    array: np.ndarray,
    colormap: str = "viridis",
    colorbar: bool = True,
    clip: bool = False,
    is_normalized: bool = False,
    title: str = "",
    as_title: bool = True,
    quantile: float = 0.995,
    img: AxesImage | None = None,
    cbar: Colorbar | None = None,
    min_val: float | None = None,
    max_val: float | None = None,
) -> tuple[AxesImage, Colorbar | None]:
    """Plot or update image."""
    if min_val is None:
        min_val = np.min(array[np.isfinite(array)])
    if clip:
        array = clip_hotspots(array, quantile)
    if not max_val and is_normalized:
        max_val = np.quantile(array[np.isfinite(array)], quantile) if quantile else np.max(array[np.isfinite(array)])
    elif not max_val:
        max_val = np.max(array[np.isfinite(array)])

    kws = {}
    if array.dtype in [np.uint, np.uint8, bool]:
        kws["interpolation"] = "nearest"

    text_color = plt.rcParams["text.color"]
    ticks, tick_labels = get_ticks_with_unit(min_val, max_val)
    if img is None:
        img = ax.imshow(array, vmax=max_val, cmap=colormap, aspect="equal", **kws)
        if colorbar:
            _, _, cbar = inset_colorbar(
                ax,
                img,
                ticks=ticks,
                ticklabels=tick_labels,
                width="76%",
                xpos=0.12,
                ypos=-0.05,
                labelcolor=text_color,
                edgecolor=text_color,
            )
            cbar.mappable.set_clim(min_val, max_val)
    else:
        img.set_array(array)
        img.set_cmap(colormap)
        img.set_clim(vmin=min_val, vmax=max_val)
        if cbar:
            cbar.mappable.set_clim(min_val, max_val)
            cbar.set_ticks(ticks)
            cbar.set_ticklabels(tick_labels)
    if as_title:
        ax.set_title(title, color=text_color, fontsize=16)
    else:
        add_label(
            ax,
            title,
            x=0.01,
            y=0.95,
            color="k",
            size=16,
            bbox={"boxstyle": "square", "facecolor": "white", "alpha": 0.75, "lw": 0},
        )
    return img, cbar


def _plot_line(
    x: np.ndarray,
    y: np.ndarray,
    marker: float | None = None,
    title: str = "",
    x_label: str = "",
    y_label: str = "",
    figsize: tuple[float, float] = (6, 6),
    marker_color: str = "r",
    border_color: str | None = None,
    title_color: str | None = None,
) -> tuple[plt.Figure, plt.Axes]:
    """Plot line."""
    fig, ax = plt.subplots(figsize=figsize)
    ax.plot(x, y)
    if marker:
        ax.axvline(marker, color=marker_color, linestyle="--", linewidth=2)
    ax.set_title(title, weight="bold", fontsize=16)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_xlim(x[0], x[-1])
    set_tick_fmt(ax, use_offset=False)
    if title_color:
        ax.title.set_color(title_color)
    if border_color:
        fig.patch.set_linewidth(2)
        fig.patch.set_edgecolor(border_color)
    return fig, ax


def plot_multiple_images(
    data: dict[str, np.ndarray],
    cmap: str | dict[str, str] = "viridis",
    style: str = "dark_background",
    title: str = "",
    figsize: tuple[int, int] = (10, 10),
    one_row: bool = False,
    n_cols: int = 0,
    highlight: str | None = None,
) -> Image:
    """Plot a pair of images."""
    from koyo.mosaic import plot_mosaic

    if one_row:
        n_cols = len(data)
    return plot_mosaic(
        data,
        colormap=cmap,
        style=style,
        figsize=figsize,
        title=title,
        n_cols=n_cols,
        highlight=highlight,
    )


def plot_correlation(
    correlation: pd.DataFrame,
    cmap: str = "coolwarm",
    vmin: float = -1,
    vmax: float = 1,
    figsize: tuple = (16, 16),
    x_label: str = "",
    y_label: str = "",
    annot: bool = False,
    tree: bool = False,
) -> plt.Figure:
    """Plot correlation matrix."""
    import seaborn as sns

    annot_kws = {}
    if annot:
        annot_kws = {"annot": True, "annot_kws": {"size": 10, "weight": "bold"}, "fmt": ".2f"}
    tree_kws = {}
    if tree:
        tree_kws = {"tree_kws": {"linewidths": 2}}

    fig = sns.clustermap(correlation.T, cmap=cmap, vmin=vmin, vmax=vmax, figsize=figsize, **annot_kws, **tree_kws)
    fig.ax_heatmap.tick_params(labelsize=12)
    fig.ax_heatmap.set_xlabel(x_label, fontsize=16)
    fig.ax_heatmap.set_ylabel(y_label, fontsize=16)
    fig.ax_heatmap.set_xticklabels(fig.ax_heatmap.get_xticklabels(), rotation=45, ha="right")
    fig.ax_heatmap.yaxis.set_tick_params(rotation=0)
    cbar = fig.ax_heatmap.collections[0].colorbar
    cbar.ax.tick_params(labelsize=10)
    return fig.fig



"""Override click group to enable ordering."""

from __future__ import annotations

import glob
import os
import sys
import traceback
import typing as ty
from ast import literal_eval
from pathlib import Path

import click
from loguru import logger
from natsort import natsorted

from koyo.typing import PathLike

ALLOW_EXTRA_ARGS = {"help_option_names": ["-h", "--help"], "ignore_unknown_options": True, "allow_extra_args": True}


def expand_data_dirs(input_dir: str) -> list[str]:
    """Expand data directory."""
    if "*" in input_dir:
        return list(glob.glob(input_dir))
    return [input_dir]


def parse_paths(input_dirs: list[PathLike], absolute: bool = False, sort: bool = False) -> list[str]:
    """Parse list/tuple of paths and check whether any of them have glob-like pattern."""
    result = []
    for path in input_dirs:
        path = str(path)
        # clean-up paths
        if path.endswith("/"):
            path = path[0:-1]
        path = expand_data_dirs(path)
        if isinstance(path, list):
            result.extend(path)
    if absolute:
        result = [os.path.abspath(path) for path in result]
    if result and sort:
        return natsorted(result)
    return result


def cli_parse_paths(ctx, param, value) -> list[str]:
    """Parse paths."""
    return parse_paths(value)


def cli_parse_paths_sort(ctx, param, value) -> list[str]:
    """Parse paths."""
    return parse_paths(value, sort=True)


def cli_parse_paths_sort_auto_glob(ctx, param, value) -> list[str]:
    """Parse paths."""
    if len(value) == 1:
        if Path(value[0]).is_dir():
            value = [f"{value[0]}/*"]
    return parse_paths(value, sort=True)


def with_plugins(plugins, **kwargs):
    """
    A decorator to register external CLI commands to an instance of
    `click.Group()`.

    Parameters
    ----------
    plugins : iter
        An iterable producing one `pkg_resources.EntryPoint()` per iteration.
    kwargs : **kwargs, optional
        Additional keyword arguments for instantiating `click.Group()`.

    Returns
    -------
    click.Group()
    """

    def decorator(group):
        if not isinstance(group, click.Group):
            raise TypeError("Plugins can only be attached to an instance of click.Group()")

        for entry_point in plugins or ():
            try:
                group.add_command(entry_point.load(), **kwargs)
            except Exception:
                # Catch this so a busted plugin doesn't take down the CLI.
                # Handled by registering a dummy command that does nothing
                # other than explain the error.
                group.add_command(BrokenCommand(entry_point.name))

        return group

    return decorator


class BrokenCommand(click.Command):
    """
    Rather than completely crash the CLI when a broken plugin is loaded, this
    class provides a modified help message informing the user that the plugin is
    broken and they should contact the owner.  If the user executes the plugin
    or specifies `--help` a traceback is reported showing the exception the
    plugin loader encountered.
    """

    def __init__(self, name):
        """Define the special help messages after instantiating a `click.Command()`."""
        click.Command.__init__(self, name)

        util_name = os.path.basename((sys.argv and sys.argv[0]) or __file__)

        if os.environ.get("CLICK_PLUGINS_HONESTLY"):  # pragma no cover
            icon = "\U0001f4a9"
        else:
            icon = "\u2020"

        self.help = (
            "\nWarning: entry point could not be loaded. Contact its author for help.\n\n\b\n" + traceback.format_exc()
        )
        self.short_help = icon + f" Warning: could not load plugin. See `{util_name} {self.name} --help`."

    def invoke(self, ctx):
        """Print the traceback instead of doing nothing."""
        click.echo(self.help, color=ctx.color)
        ctx.exit(1)

    def parse_args(self, ctx, args):
        return args


_verbosity_options = [
    click.option(
        "--verbose",
        "-v",
        "verbosity",
        default=1,
        count=True,
        help="Verbose output. This is additive flag so `-vvv` will print `INFO` messages and -vvvv will print `DEBUG`"
        " information.",
    ),
    click.option("--quiet", "-q", "verbosity", flag_value=0, help="Minimal output"),
    click.option("--debug", "verbosity", flag_value=5, help="Maximum output"),
]


def verbosity_options(func: ty.Callable) -> ty.Callable:
    """Verbosity options."""
    for option in reversed(_verbosity_options):
        func = option(func)
    return func


def dev_options(func: ty.Callable) -> ty.Callable:
    """Setup dev options."""
    from koyo.system import IS_PYINSTALLER

    if not IS_PYINSTALLER:
        func = click.option(
            "--dev",
            help="Flat to indicate that CLI should run in development mode and catch all errors.",
            default=False,
            is_flag=True,
            show_default=True,
        )(func)
    return func


def repr_filelist(filelist: ty.Sequence[PathLike]) -> str:
    """Pretty iterable..."""
    res = [Path(file).name for file in filelist]
    return "; ".join(res)


def repr_iterable(iterable: ty.Sequence[ty.Any]):
    """Pretty iterable..."""
    if iterable is None:
        return iterable
    if isinstance(iterable, ty.Sequence):
        n = len(iterable)
        if n > 10:
            return f"{n} items..."
        return iterable


def append(table: list[tuple[str, str, str]], name: str = "", param: str = "", value: ty.Any = "") -> None:
    """Append to table."""
    if isinstance(value, Path):
        value = str(value)
    if isinstance(value, str) and len(value) > 120:
        value = "..." + str(value)[-120:]
    elif isinstance(value, (list, tuple)):
        value = repr_iterable(value)
    elif isinstance(value, bool):
        value = str(value)
    elif value is None:
        value = "<no value>"
    table.append((name, param, value))


def format_value(description: str, args: str, value: ty.Any) -> list[tuple[str, str, str]]:
    """Format value."""
    res = []
    # lists should be printed as multiple rows
    if isinstance(value, (tuple, list)):
        if len(value) > 0:
            append(res, description, args, value[0])
            for v in value[1::]:
                append(res, "", "", v)
        else:
            append(res, description, args, "<no value>")
    # dicts should be printed as multiple rows
    elif isinstance(value, dict):
        if len(value) > 0:
            for i, (k, v) in enumerate(value.items()):
                if i == 0:
                    append(res, description, args, f"{k}={v}")
                else:
                    append(res, "", "", f"{k}={v}")
        else:
            append(res, description, args, "<no value>")
    else:
        append(res, description, args, value)
    return res


def get_args_from_option(option: ty.Callable) -> str:
    """Return argument information from option."""
    if not hasattr(option, "__closure__"):
        raise ValueError(f"Option {option} does not have closure.")
    closure = option.__closure__
    for cell in closure:
        if isinstance(cell.cell_contents, tuple):
            break
    ret = ""
    for value in cell.cell_contents:
        if isinstance(value, str) and value.startswith("-"):
            ret += value
    # check if we can split the arguments
    if "--" in ret:
        rets = ret.split("--")
        rets = [r if r.startswith("-") else f"--{r}" for r in rets]
        ret = "/".join(rets)
    return ret


class Parameter:
    """Parameter object."""

    __slots__ = ["args", "description", "value"]

    def __init__(self, description: str, args: str | ty.Callable | None, value: ty.Any | None = None):
        self.description = description
        self.args = get_args_from_option(args) if callable(args) else args
        self.value = value

    def with_value(self, value: ty.Any, description: str | None = None, args: str | None = None) -> Parameter:
        """Set the value of the parameter and return self."""
        return Parameter(description or self.description, args or self.args, value)

    def to_list(self) -> list[tuple[str, str, str]]:
        """Return formatted version of the value."""
        if self.value is not None:
            return format_value(self.description, self.args, self.value)
        return [(self.description, self.args, "<no value>")]


class Parameters:
    """Parameters class."""

    @classmethod
    def p(cls, description: str, args: str, value: ty.Any):
        """Return Parameter instance."""
        return Parameter(description, args, value)


def print_parameters(*parameters: Parameter, log: bool = True, silent: bool = False):
    """Print parameters as table."""
    from tabulate import tabulate

    if silent:
        return

    table = []
    for param in parameters:
        table.extend(param.to_list())
    table = tabulate(table, headers=["Name", "Parameter", "Value"], tablefmt="github")
    if not log:
        print(table)
    else:
        for line in table.split("\n"):
            logger.info(line)


def exception_msg(msg: str):
    """Display error message."""
    logger.exception(msg)


def error_msg(msg: str):
    """Display error message."""
    logger.error(msg)


def warning_msg(msg: str):
    """Display warning message."""
    logger.warning(msg)


def info_msg(msg: str):
    """Display info message."""
    logger.info(msg)


def success_msg(msg: str):
    """Display success message."""
    logger.info(msg)


def expand_dirs(input_dir: str) -> ty.Sequence[str]:
    """Expand data directory."""
    if "*" in str(input_dir):
        return glob.glob(input_dir)
    elif isinstance(input_dir, (str, Path)):
        return [str(input_dir)]
    elif isinstance(input_dir, ty.Sequence):
        return [str(path) for path in input_dir]
    return input_dir


# noinspection PyUnusedLocal
def arg_parse_path(ctx, param, value) -> list[Path]:
    """Split arguments."""
    if value is None:
        return []
    res: list[str] = []
    if isinstance(value, (str, Path)):
        value = [value]
    for path in value:
        path = str(path).replace("'", "").replace('"', "")
        if path.endswith("/"):
            path = path[:-1]
        res.extend(expand_dirs(path))
    return [Path(path).absolute() for path in res]


# noinspection PyUnusedLocal
def arg_split_str(ctx, param, value):
    """Split arguments."""
    if value is None:
        return None
    args = [arg.strip() for arg in value.split(",")]
    return args


# noinspection PyUnusedLocal
def arg_split_float(ctx, param, value):
    """Split arguments."""
    if value is None:
        return None
    args = [float(arg.strip()) for arg in value.split(",")]
    return args


# noinspection PyUnusedLocal
def arg_split_int(ctx, param, value):
    """Split arguments."""
    if value is None:
        return None
    args = [int(arg.strip()) for arg in value.split(",")]
    return args


# noinspection PyUnusedLocal
def arg_parse_env(ctx, param, value) -> tuple[str | None, ty.Any | None]:
    """Split argument into environment variables."""
    if value is None:
        return None, None
    return [v.split("=", maxsplit=1) for v in value]


def parse_str_framelist(framelist: str) -> list[int]:
    """Parse list provided in string format."""
    if framelist is None:
        return []
    framelist = framelist.replace(" ", "")
    _framelist = framelist.split(",")

    out_framelist = []
    for frame in _framelist:
        _frame = frame.split(":")
        if len(_frame) == 1:
            out_framelist.append(int(_frame[0]))
        elif len(_frame) in [2, 3]:
            start, end, step = int(_frame[0]), int(_frame[1]), 1
            if len(_frame) == 3:
                step = int(_frame[2])
            out_framelist.extend(list(range(start, end, step)))
        else:
            print(f"Skipped {frame} - could not parse it")
    return out_framelist


# noinspection PyUnusedLocal
def arg_parse_framelist(ctx, param, value: str):
    """Parse framelist."""
    if value is None:
        return None
    return parse_str_framelist(value)


def arg_parse_framelist_multi(ctx, param, value: tuple[str]):
    """Parse framelist."""
    if value is None:
        return None
    return [parse_str_framelist(v) for v in value]


# noinspection PyUnusedLocal
def parse_values(ctx, param, value: str):
    """Parse values and return them as list of potential options."""
    from ast import literal_eval

    if isinstance(value, str):
        return [literal_eval(val) for val in value.split(",")]
    return [literal_eval(val) for val in value]


def timed_iterator(
    iterable: ty.Iterable,
    text: str = "Task executed in",
    func: ty.Callable = info_msg,
    silent: bool = False,
    is_filename: bool = False,
):
    """Timed iterable that yields value and prints amount of time spent on said iterable."""
    from koyo.timer import format_human_time_s, measure_time

    total = 0
    iterable = list(iterable)
    n_tasks = len(iterable)
    for i, item in enumerate(iterable, start=1):
        with measure_time() as timer:
            yield item
        if not silent:
            # update timing
            execution_time = timer()
            total += execution_time
            avg = total / i
            remaining = avg * (n_tasks - i)
            _text = text if not is_filename else text + f" ({os.path.basename(item)})"
            avg = f"avg={format_human_time_s(avg)}; "
            rem = f"rem={format_human_time_s(remaining)}; " if remaining > 0 else ""
            tot = f"tot={format_human_time_s(total)}"
            func(f"[{i}/{n_tasks}] {_text} {format_human_time_s(execution_time)} [{avg}{rem}{tot}]")


def parse_arg(arg: str, key: str) -> tuple[str, ty.Any]:
    """Parse argument."""
    try:
        if key:
            arg = arg.split(key)[1]
        name, value = arg.split("=", maxsplit=1)
        # try parsing value - it will fail if string value was provided
        try:
            value = literal_eval(value)
        # try wrapping value, so it can be evaluated as string literal
        except ValueError:
            value = f'"{value}"'
            value = literal_eval(value)
        except SyntaxError:
            pass
        return name, value
    except ValueError:
        raise ValueError(f"Could not parse argument {arg}")


def parse_extra_args(extra_args: tuple[str, ...] | None) -> dict[str, ty.Any]:
    """Arguments"""
    kwargs: dict[str, ty.Any] = {}
    if extra_args is None:
        return kwargs
    for arg in extra_args:
        if "=" not in arg:
            info_msg(f"Skipping argument {arg} as it does not contain '='")
            continue
        name, value = parse_arg(arg, "")
        if name in kwargs:
            if name in kwargs and not isinstance(kwargs[name], list):
                kwargs[name] = [kwargs[name]]
            if isinstance(kwargs[name], list):
                kwargs[name].append(value)
            elif isinstance(kwargs[name], (str, int, float, bool)):
                kwargs[name] = value
            else:
                kwargs[name] = [kwargs[name], value]
        else:
            kwargs[name] = value
    return kwargs


def parse_args_with_keys(
    extra_args: tuple[str] | None, fields: tuple[str, ...], clean: bool = False
) -> dict[str, ty.Any] | tuple[dict[str, ty.Any], tuple[str]]:
    extra_kwargs = {}
    if extra_args is None:
        if clean:
            return extra_kwargs, extra_args
        return extra_kwargs
    extra_args_ = []
    for arg in extra_args:
        if any(arg.startswith(k) for k in fields):
            for k in fields:
                if arg.startswith(k):
                    name, value = parse_arg(arg, k)
                    extra_kwargs[name] = value
                    break
        else:
            extra_args_.append(arg)
    if clean:
        return extra_kwargs, extra_args_
    return extra_kwargs


def parse_fig_args(
    extra_args: tuple[str] | None, clean: bool = False
) -> dict[str, ty.Any] | tuple[dict[str, ty.Any], tuple[str]]:
    """Parse extra parameters."""
    return parse_args_with_keys(extra_args, ("--fig:", "--f:"), clean)


@ty.overload
def parse_env_args(extra_args: tuple[str, ...] | None, clean: bool = False) -> dict[str, ty.Any]:
    """Parse extra environment variables."""
    ...


@ty.overload
def parse_env_args(extra_args: tuple[str, ...] | None, clean: bool = False) -> tuple[dict[str, ty.Any], tuple[str]]:
    """Parse extra environment variables."""
    ...


def parse_env_args(
    extra_args: tuple[str, ...] | None, clean: bool = False
) -> dict[str, ty.Any] | tuple[dict[str, ty.Any], tuple[str]]:
    """Parse extra environment variables."""
    env_kwargs = {}
    if extra_args is None:
        if clean:
            return env_kwargs, extra_args
        return env_kwargs

    parsed_extra_args = []
    for arg in extra_args:
        if arg.startswith("--env:"):
            name, value = parse_arg(arg, "--env:")
            env_kwargs[name] = value
        else:
            parsed_extra_args.append(arg)
    if clean:
        return env_kwargs, parsed_extra_args
    return env_kwargs


def set_env_args(**kwargs: ty.Any) -> None:
    """Set environment variables."""
    for name, value in kwargs.items():
        os.environ[name] = str(value)
        logger.info(f"Set environment variable: {name}={value}")


def set_env_args_from_tuples(env_vars: tuple[str | None, str | None]) -> None:
    """Set environment variables."""
    for name, value in env_vars:
        if name is None or value is None:
            continue
        os.environ[name] = str(value)
        logger.info(f"Set environment variable: {name}={value}")


def filter_kwargs(*allowed: str, **kwargs: ty.Any) -> dict[str, ty.Any]:
    """Filter allowed kwargs."""
    return {key: value for key, value in kwargs.items() if key in allowed}


def exit_with_error(skip_error: bool = False) -> int | None:
    """Skip error or exit with error."""
    import sys

    if skip_error:
        return sys.exit()
    return sys.exit(1)


def select_from_list(
    item_list: list[ty.Any],
    text: str = "Please select one of the options from the list",
    auto_select: str = "off",
    default: int = -1,
) -> int:
    """Select file from the list. The list is assumed to be sorted in descending order of creation time meaning that
    oldest files should be first and newest should be last.
    """
    if item_list:
        if len(item_list) == 1:
            return 0
        elif auto_select.lower() == "off":
            choice = click.prompt(text, type=click.INT, default=default)
            return choice
        elif auto_select.lower() == "newest":
            return len(item_list) - 1
        elif auto_select.lower() == "oldest":
            return 0
    return default


# Commonly used click options
extra_args_ = click.argument("extra_args", nargs=-1, type=click.UNPROCESSED)
env_var_ = click.option(
    "-e",
    "--env",
    type=str,
    multiple=True,
    help="Environment variables to set. Values should be set as KEY=VALUE (e.g. --env KEY=VALUE).",
    callback=arg_parse_env,
)



"""Test color functions"""
import pytest
from koyo.color import get_random_hex_color, hex_to_rgb_1, hex_to_rgb_255, rgb_255_to_1


def test_convert_rgb_255_to_1():
    expected_color = [1.0, 1.0, 1.0]

    returned_color = rgb_255_to_1([255, 255, 255])
    assert set(expected_color) == set(returned_color)

    returned_color = rgb_255_to_1("[255, 255, 255]")
    assert set(expected_color) == set(returned_color)

    with pytest.raises(ValueError):
        rgb_255_to_1([255, 255, 255, 255, 255])
    with pytest.raises(ValueError):
        rgb_255_to_1([255, 255, 255, -1])
    with pytest.raises(ValueError):
        rgb_255_to_1([255, 255, 255, 256])


@pytest.mark.parametrize("hex_color, rgb_expected", (["#000000", [0, 0, 0]], ["#FFFFFF", [1, 1, 1]]))
def test_convert_hex_to_rgb_1(hex_color, rgb_expected):
    rgb_result = hex_to_rgb_1(hex_color)
    assert rgb_result == rgb_expected


@pytest.mark.parametrize("hex_color, rgb_expected", (["#000000", [0, 0, 0]], ["#FFFFFF", [255, 255, 255]]))
def test_convert_hex_to_rgb_255(hex_color, rgb_expected):
    rgb_result = hex_to_rgb_255(hex_color)
    assert rgb_result == rgb_expected


def test_get_random_hex_color():
    color = get_random_hex_color()
    assert color.startswith("#")
    color_rgb = hex_to_rgb_255(color)
    assert all(value <= 255 for value in color_rgb)



"""Statistics functions."""

from __future__ import annotations

import numpy as np
import pandas as pd
import scipy.stats as ss


def rank_features(df: pd.DataFrame) -> np.ndarray:
    """Rank features."""
    return np.apply_along_axis(ss.rankdata, 0, df)


def pairwise_spearmanr(
    array_one: np.ndarray | pd.DataFrame, array_two: np.ndarray | pd.DataFrame | None = None, ranked: bool = False
) -> np.ndarray | pd.DataFrame:
    """Compute pairwise Spearman rank correlation between columns of two arrays."""
    if array_two is None:
        array_two = array_one

    columns_one = None
    if isinstance(array_one, pd.DataFrame):
        columns_one = array_one.columns
        array_one = array_one.values
    columns_two = None
    if isinstance(array_two, pd.DataFrame):
        columns_two = array_two.columns
        array_two = array_two.values
    if not ranked:
        array_one = np.apply_along_axis(ss.rankdata, 0, array_one)
        array_two = np.apply_along_axis(ss.rankdata, 0, array_two)
    am = array_one - np.mean(array_one, axis=0, keepdims=True)
    bm = array_two - np.mean(array_two, axis=0, keepdims=True)

    correlation_matrix = (
        am.T
        @ bm
        / (np.sqrt(np.sum(am**2, axis=0, keepdims=True) + 1).T * np.sqrt(np.sum(bm**2, axis=0, keepdims=True)) + 1)
    )
    if columns_one is not None and columns_two is not None:
        correlation_matrix = pd.DataFrame(correlation_matrix, index=columns_one, columns=columns_two)
    return correlation_matrix


def pairwise_pearsonr(array_one: np.ndarray, array_two: np.ndarray | pd.DataFrame | None = None) -> np.ndarray:
    """Compute pairwise Pearson correlation between columns of two arrays.

    Parameters
    ----------
    - array_one: numpy array of shape (N, M1)
    - array_two: numpy array of shape (N, M2)

    Returns
    -------
    - correlation_matrix: numpy array of shape (M1, M2) with Pearson correlation coefficients
    """

    def _get_stats(array):
        mean = array.mean(axis=0)
        std = array.std(axis=0)
        std[std == 0] = 1
        return (array - mean) / std

    if array_two is None:
        array_two = array_one

    columns_one = None
    if isinstance(array_one, pd.DataFrame):
        columns_one = array_one.columns
        array_one = array_one.values
    columns_two = None
    if isinstance(array_two, pd.DataFrame):
        columns_two = array_two.columns
        array_two = array_two.values

    array1_normalized = _get_stats(array_one)
    array2_normalized = _get_stats(array_two)

    # Compute the correlation matrix as dot product
    correlation_matrix = np.dot(array1_normalized.T, array2_normalized) / array_one.shape[0]
    if columns_one is not None and columns_two is not None:
        correlation_matrix = pd.DataFrame(correlation_matrix, index=columns_one, columns=columns_two)
    return correlation_matrix


def get_significant_correlations(
    correlation_matrix: np.ndarray | pd.DataFrame, threshold: float = 0.5
) -> list[tuple[int, int, float]]:
    """Get significant correlations."""
    if isinstance(correlation_matrix, pd.DataFrame):
        correlation_matrix = correlation_matrix.values

    pairs = [
        (i, j, float(correlation_matrix[i, j]))
        for i in range(correlation_matrix.shape[0])
        for j in range(correlation_matrix.shape[1])
        if abs(correlation_matrix[i, j]) > threshold
    ]
    pairs = sorted(pairs, key=lambda x: x[2])
    return pairs


def get_n_significant_correlations(
    correlation_matrix: np.ndarray | pd.DataFrame, n: int
) -> tuple[list[tuple[int, int, float]], list[tuple[int, int, float]]]:
    """Get the n highest and lowest correlations."""
    if isinstance(correlation_matrix, pd.DataFrame):
        correlation_matrix = correlation_matrix.values

    pairs = []
    for i in range(correlation_matrix.shape[0]):
        for j in range(i + 1, correlation_matrix.shape[1]):
            pairs.append((i, j, float(correlation_matrix[i, j])))
    pairs = sorted(pairs, key=lambda x: x[2])
    anti_correlated = pairs[:n]
    correlated = pairs[-n:][::-1]
    return anti_correlated, correlated



import pytest
from koyo.json import read_json_data, read_json_gzip, write_json_data, write_json_gzip


@pytest.mark.parametrize("indent", (0, 4))
@pytest.mark.parametrize("data", ({"A": 1}, [0, 1, 2], ["A", "B", 3, 4]))
def test_read_write(tmpdir_factory, data, indent):
    path = str(tmpdir_factory.mktemp("json") / "test.json")
    write_json_data(path, data, indent=indent)

    _data = read_json_data(path)
    assert _data == data


@pytest.mark.parametrize("data", ({"A": 1}, [0, 1, 2], ["A", "B", 3, 4]))
def test_read_write_gzip(tmpdir_factory, data):
    path = str(tmpdir_factory.mktemp("json") / "test.json")
    write_json_gzip(path, data)

    _data = read_json_gzip(path)
    assert _data == data


def test_check_gzip_size(tmpdir_factory):
    path_json = str(tmpdir_factory.mktemp("json") / "test.json")
    path_gz = str(tmpdir_factory.mktemp("json") / "test-gz.json")
    # create large dictionary
    data = {str(i): i for i in range(1000)}
    path_json = write_json_data(path_json, data)
    path_gz = write_json_gzip(path_gz, data)
    assert path_json.stat().st_size > path_gz.stat().st_size



"""All utility functions that deal with JSON files."""

import typing as ty
from pathlib import Path

import numpy as np

from koyo.typing import PathLike
from koyo.utilities import _remove_duplicates_from_dict

try:
    import ujson as json
except ImportError:
    import json


def default(o: ty.Any) -> ty.Any:
    """Fixes writing of objects containing numpy dtypes.

    Parameters
    ----------
    o : any
        object to be serialized

    Returns
    -------
    int / float
        converted object

    Raises
    ------
    TypeError
    """
    if isinstance(o, (np.int64, np.int32, np.int16, np.integer)):
        return int(o)
    elif isinstance(o, (np.float64, np.float32, np.float16, np.floating)):
        return float(o)
    elif isinstance(o, np.ndarray):
        return o.tolist()
    elif isinstance(o, np.bool_):
        return bool(o)
    elif isinstance(o, Path):
        return str(o)
    raise TypeError("Could not convert {} of type {}".format(*o), type(o))


def read_json(filepath: PathLike) -> ty.Any:
    """Read JSON data and metadata.

    Parameters
    ----------
    filepath : PathLike
        path to the JSON file

    Returns
    -------
    loaded json data
    """
    with open(filepath) as f_ptr:
        obj = json.load(f_ptr)
    return obj


def write_json(filepath: PathLike, obj, indent: int = 4, check_existing: bool = False, compress: bool = False) -> Path:
    """Write data to JSON file.

    Parameters
    ----------
    filepath : PathLike
        path to JSON file
    obj : any
        object containing data
    indent : int, optional
        number of spaces to indent by, by default 4
    check_existing : bool, optional
        if True, existing JSON file data will be merged with the new data object
    compress : bool, optional
    """
    kws = {} if not compress else {"separators": (",", ":")}
    indent = 1 if compress else indent

    filepath = Path(filepath)
    if not check_existing or not filepath.exists():
        with open(filepath, "w") as f_ptr:
            json.dump(obj, f_ptr, indent=indent, default=default, **kws)
    else:
        with open(filepath, "r+") as f_ptr:
            data = json.load(f_ptr)

            if isinstance(data, list):
                data.extend(obj)
            elif isinstance(data, dict):
                data.update(obj)

            # remove duplicates
            data = _remove_duplicates_from_dict(data)

            f_ptr.seek(0)  # rewind
            json.dump(data, f_ptr, indent=indent, default=default, **kws)
            f_ptr.truncate()
    return filepath


def write_json_gzip(filepath: PathLike, obj: ty.Any) -> Path:
    """Write gzip compressed JSON data."""
    import gzip

    filepath = Path(filepath)
    if not filepath.with_suffix(".gz"):
        filepath = filepath.with_suffix(".gz")
    with gzip.open(filepath, "w") as f_ptr:
        f_ptr.write(json.dumps(obj, default=default).encode("utf-8"))
    return filepath


def read_json_gzip(filepath: PathLike) -> ty.Any:
    """Read gzip compressed JSON data."""
    import gzip

    with gzip.open(filepath, "r") as f_ptr:
        return json.loads(f_ptr.read().decode("utf-8"))


# compatibility
read_json_data = read_json
write_json_data = write_json



import numpy as np
import pytest
from koyo.sparse import get_array_statistics
from scipy import sparse


@pytest.mark.parametrize(
    "array",
    (
        np.zeros((10, 10)),
        np.ones((10, 10)),
        sparse.csr_matrix((3, 4), dtype=np.int8),
        sparse.csc_matrix((3, 4), dtype=np.int8),
    ),
)
def test_get_array_statistics(array):
    result = get_array_statistics(array)
    assert isinstance(result, str)
    assert "Sparsity" in result



"""PowerPoint mixin class."""

from __future__ import annotations

import io
import typing as ty
from contextlib import contextmanager
from enum import IntEnum
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from loguru import logger

from koyo.typing import PathLike
from koyo.utilities import is_installed

HAS_PPTX = is_installed("pptx")

if ty.TYPE_CHECKING:
    from PIL import Image

    try:
        from pptx.presentation import Presentation
    except ImportError:
        Presentation = None  # type: ignore[assignment,misc]


class SlideLayout(IntEnum):
    """Slide layout options."""

    TITLE = 0
    TITLE_AND_CONTENT = 1
    SECTION_HEADER = 2
    TITLE_AND_TWO_CONTENT = 3
    TITLE_AND_TWO_CONTENT_WITH_HEADER = 4
    TITLE_AND_BLANK = 5
    BLANK = 6
    HEADER_AND_TWO_CONTENT = 7
    PICTURE_AND_CAPTION = 8


class PPTXMixin:
    """Mixin class to help export figures to PPTX files."""

    _pptx = None
    as_pptx: bool = False

    @property
    def pptx_filename(self) -> Path:
        """Return the PDF filename."""
        raise NotImplementedError("Must implement method")

    @property
    def pptx(self) -> Presentation | None:
        """Return PDF figure."""
        if self._pptx is None and self.as_pptx:
            self._pptx = self._make_pptx(self.pptx_filename)
        return self._pptx

    @pptx.setter
    def pptx(self, value: Presentation | None) -> None:
        """Set PDF figure."""
        if self._pptx is not None:
            self._pptx.save(self.pptx_filename)  # type: ignore[arg-type]
        self._pptx = value

    @staticmethod
    def _make_pptx(filename: PathLike) -> Presentation:
        """Make PDF."""
        try:
            from pptx import Presentation
            from pptx.util import Inches
        except ImportError:
            raise ImportError("pptx is not installed. Please install it using `pip install python-pptx`.") from None

        pptx = Presentation()
        pptx.slide_width = Inches(16)
        pptx.slide_height = Inches(9)
        pptx._filename = filename  # type: ignore[attr-defined]
        return pptx

    @contextmanager
    def _export_pptx_figures(self, filename: PathLike | None = None) -> ty.Generator[Presentation | None, None, None]:
        """Export figures."""
        import matplotlib

        matplotlib.use("agg")

        pptx, reset = None, False
        if self.as_pptx:
            if filename:
                pptx = self._make_pptx(filename)
            else:
                pptx = self.pptx
                reset = True
        try:
            yield pptx
        except Exception as e:
            logger.exception(f"Error exporting to PPTX: {e}")
        finally:
            if self.as_pptx:
                self._save_pptx(pptx, filename, reset)  # type: ignore[arg-type]

    def _add_title_to_pptx(self, title: str, pptx: Presentation | None = None) -> None:
        """Add title page to the slide deck."""
        pptx = pptx or self.pptx
        add_title_to_pptx(pptx, title)  # type: ignore[arg-type]

    def _add_content_to_pptx(self, content: str, title: str = "", pptx: Presentation | None = None) -> None:
        """Add title page to the slide deck."""
        pptx = pptx or self.pptx
        add_content_to_pptx(pptx, content, title)  # type: ignore[arg-type]

    def _add_mpl_figure_to_pptx(
        self,
        filename: PathLike,
        fig: plt.Figure,
        face_color: str | np.ndarray | None = None,
        bbox_inches: str | None = "tight",
        dpi: int = 150,
        override: bool = False,
        if_empty: str = "warn",
        close: bool = False,
        title: str = "",
        pptx: Presentation | None = None,
        **kwargs: ty.Any,
    ) -> None:
        """Export figure to file."""
        if fig is None:
            self._inform_on_empty(if_empty)
            return

        pptx = pptx or self.pptx
        add_mpl_figure_to_pptx(
            pptx, filename, fig, face_color, bbox_inches, dpi, override, close=close, title=title, **kwargs
        )

    def _add_pil_image_to_pptx(
        self,
        filename: PathLike,
        image: Image,
        dpi: int = 150,
        fmt: str = "JPEG",
        override: bool = False,
        close: bool = False,
        title: str = "",
        pptx: Presentation | None = None,
        **kwargs: ty.Any,
    ) -> None:
        """Add PIL image to pptx."""
        pptx = pptx or self.pptx
        add_pil_image_to_pptx(
            pptx, filename, image, dpi, fmt=fmt, override=override, close=close, title=title, **kwargs
        )

    def _save_pptx(self, pptx: Presentation, filename: PathLike | None = None, reset: bool = False) -> None:
        """Save PPTX."""
        if hasattr(pptx, "_filename"):
            filename = pptx._filename
        else:
            filename = filename or self.pptx_filename
        pptx.save(filename)  # type: ignore[arg-type]
        logger.trace(f"Saved PPTX to {filename} with {len(pptx.slides)} slides")
        if reset:
            self._pptx = None

    @staticmethod
    def _inform_on_empty(if_empty: str = "warn") -> None:
        """Inform the user if the figure was empty."""
        if if_empty == "none":
            return
        elif if_empty == "warn":
            logger.warning("Figure was empty")
        elif if_empty == "raise":
            raise ValueError("Figure was empty")


def add_title_to_pptx(pptx: Presentation, title: str) -> None:
    """Add title to the slide."""
    from pptx.util import Cm

    slide = pptx.slides.add_slide(pptx.slide_layouts[SlideLayout.TITLE])
    count = title.count("\n") + 1
    height = slide.shapes.title.height * count
    slide.shapes.title.text = title
    # adjust positions
    slide.shapes.title.left = Cm(0)
    slide.shapes.title.width = pptx.slide_width
    slide.shapes.title.height = height


def add_content_to_pptx(pptx: Presentation, content: str, title: str = "") -> None:
    """Add title to the slide."""
    from pptx.util import Cm

    slide = pptx.slides.add_slide(pptx.slide_layouts[SlideLayout.TITLE_AND_CONTENT])
    count = title.count("\n") + 1
    height = slide.shapes.title.height * count
    slide.shapes.title.text = title
    slide.placeholders[1].text = content
    # adjust positions
    slide.shapes.title.left = Cm(0)
    slide.shapes.title.width = pptx.slide_width
    slide.shapes.title.height = height


def add_mpl_figure_to_pptx(
    pptx: Presentation | None,
    filename: PathLike,
    fig: plt.Figure,
    face_color: str | np.ndarray | None = None,
    bbox_inches: str | None = "tight",
    dpi: int = 150,
    override: bool = False,
    close: bool = False,
    title: str = "",
    format: str = "jpg",
    **kwargs: ty.Any,
) -> None:
    """Export figure to file."""
    face_color = face_color if face_color is not None else fig.get_facecolor()
    if pptx is not None:
        with io.BytesIO() as image_stream:
            fig.savefig(image_stream, dpi=dpi, facecolor=face_color, bbox_inches=bbox_inches, format=format, **kwargs)
            slide, left, top = _insert_slide(pptx, title=title)
            slide.shapes.add_picture(image_stream, left, top)  # , width=pptx.slide_width, height=pptx.slide_height)
    else:
        if override or not Path(filename).exists():
            fig.savefig(filename, dpi=dpi, facecolor=face_color, bbox_inches=bbox_inches, **kwargs)
    if close:
        plt.close(fig)


def add_pil_image_to_pptx(
    pptx: Presentation | None,
    filename: PathLike,
    image: Image,
    dpi: int = 150,
    fmt: str = "JPEG",
    override: bool = False,
    close: bool = False,
    title: str = "",
    **kwargs: ty.Any,
) -> None:
    """Export figure to file."""
    quality = kwargs.pop("quality", 95)
    if pptx is not None:
        if image.mode == "RGBA" and fmt.upper() in ["JPEG", "JPG"]:
            image = image.convert("RGB")
        with io.BytesIO() as image_stream:
            image.save(image_stream, fmt, quality=quality, dpi=(dpi, dpi), **kwargs)
            slide, left, top = _insert_slide(pptx, title=title)
            slide.shapes.add_picture(image_stream, left, top)  # , width=pptx.slide_width, height=pptx.slide_height)
    else:
        if override or not Path(filename).exists():
            image.save(filename, dpi=(dpi, dpi), **kwargs)
    if close:
        image.close()


def _insert_slide(pptx: Presentation, title: str = "") -> tuple[ty.Any, int, int]:
    left = top = 0
    template_index = SlideLayout.TITLE_AND_BLANK if title else SlideLayout.BLANK
    slide = pptx.slides.add_slide(pptx.slide_layouts[template_index])
    if title:
        from pptx.util import Cm, Pt

        count = title.count("\n") + 1

        slide.shapes.title.text_frame.paragraphs[0].font.size = Pt(20)
        slide.shapes.title.text = title
        height = slide.shapes.title.height // 2
        height *= count
        slide.shapes.title.top = Cm(0)
        slide.shapes.title.left = Cm(0)
        slide.shapes.title.width = pptx.slide_width
        slide.shapes.title.height = height
        top = slide.shapes.title.top + height
    return slide, left, top



"""Test zarr functionality."""

import numpy as np
import pytest

from koyo.utilities import is_installed
from koyo.zarr import get_chunk_shape_along_axis, load_array_from_zip, save_array_to_zip


@pytest.mark.skipif(not is_installed("zarr"), reason="zarr not installed")
def test_get_chunk_shape_along_axis():
    """Test chunk shape retrieval."""
    array_1d = np.zeros(10)
    assert get_chunk_shape_along_axis(array_1d, 0) == (10,)

    array_2d = np.zeros((10, 20))
    assert get_chunk_shape_along_axis(array_2d, 0) == (10, 1)
    assert get_chunk_shape_along_axis(array_2d, 1) == (1, 20)


@pytest.mark.skipif(not is_installed("zarr"), reason="zarr not installed")
def test_save_load_array_to_zip(tmp_path):
    """Test saving and loading a 2D array to/from a zip file."""
    array = np.random.rand(100, 200)
    zip_path = tmp_path / "test_array.zip"

    # Save the array
    save_array_to_zip(array, str(zip_path), chunk_size=(50, 50))
    assert zip_path.exists(), "Zip file was not created"

    # Load the array back
    loaded_array = load_array_from_zip(str(zip_path))

    # Check if the loaded array matches the original
    np.testing.assert_array_equal(array, loaded_array)


@pytest.mark.skipif(not is_installed("zarr"), reason="zarr not installed")
def test_load_array_from_zip_nonexistent():
    """Test loading from a non-existent zip file."""
    with pytest.raises(FileNotFoundError):
        load_array_from_zip("non_existent.zip")



"""Test PDF utilities."""

from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pytest
from koyo.fig_mixin import FigureExporter, FigureMixin
from koyo.pptx_mixin import HAS_PPTX
from PIL import Image


class PptxPdf(FigureMixin):
    """Mixin class."""

    def __init__(self, filename: Path, as_pdf: bool = False, as_pptx: bool = False):
        self.as_pdf = as_pdf
        self.as_pptx = as_pptx
        self._filename = Path(self.get_pptx_or_pdf_filename(filename))

    @property
    def pdf_filename(self) -> Path:
        """Filename."""
        return self._filename

    @property
    def pptx_filename(self) -> Path:
        """Filename."""
        return self._filename


def test_init(tmp_path):
    tmp = Path(tmp_path)
    # init PDF
    pptx_or_pdf = PptxPdf(tmp / "test.pdf", as_pdf=True, as_pptx=True)
    with pytest.raises(ValueError):
        pptx_or_pdf._check_export()


def test_api_auto_pdf(tmp_path):
    tmp = Path(tmp_path)
    pptx_or_pdf = PptxPdf(tmp / "test", as_pdf=True, as_pptx=False)
    filename = pptx_or_pdf.get_pptx_or_pdf_filename("test")
    with pptx_or_pdf._export_figures() as obj:
        assert obj is not None
        plt.plot([1, 2, 3], [1, 2, 3])
        obj.add_or_export_mpl_figure(tmp / "test.png", plt.gcf(), title="test")
        assert not (tmp / "test.png").exists()
        array = np.random.randint(0, 255, (1000, 1000), dtype=np.uint8)
        image = Image.fromarray(array)
        obj.add_or_export_pil_image(tmp / "test.png", image, title="test")
        assert not (tmp / "test.png").exists()
        obj.add_title("Title")
    assert (tmp / filename).exists()


@pytest.mark.skipif(not HAS_PPTX, reason="pptx not installed")
def test_api_auto_pptx(tmp_path):
    tmp = Path(tmp_path)
    pptx_or_pdf = PptxPdf(tmp / "test", as_pdf=False, as_pptx=True)
    filename = pptx_or_pdf.get_pptx_or_pdf_filename("test")
    with pptx_or_pdf._export_figures() as obj:
        assert obj is not None
        plt.plot([1, 2, 3], [1, 2, 3])
        obj.add_or_export_mpl_figure(tmp / "test.png", plt.gcf(), title="test")
        assert not (tmp / "test.png").exists()
        array = np.random.randint(0, 255, (1000, 1000), dtype=np.uint8)
        image = Image.fromarray(array)
        obj.add_or_export_pil_image(tmp / "test.png", image, title="test")
        assert not (tmp / "test.png").exists()
        obj.add_title("Title")
    assert (tmp / filename).exists()


def test_api_pdf(tmp_path):
    tmp = Path(tmp_path)
    pptx_or_pdf = PptxPdf(tmp / "test.pdf", as_pdf=True, as_pptx=False)

    with pptx_or_pdf._export_figures() as obj:
        assert obj is not None
        plt.plot([1, 2, 3], [1, 2, 3])
        obj.add_or_export_mpl_figure(tmp / "test.png", plt.gcf())
        assert not (tmp / "test.png").exists()
        array = np.random.randint(0, 255, (1000, 1000), dtype=np.uint8)
        image = Image.fromarray(array)
        obj.add_or_export_pil_image(tmp / "test.png", image)
        assert not (tmp / "test.png").exists()
        obj.add_title("Title")
    assert (tmp / "test.pdf").exists()


@pytest.mark.skipif(not HAS_PPTX, reason="pptx not installed")
def test_api_pptx(tmp_path):
    tmp = Path(tmp_path)
    pptx_or_pdf = PptxPdf(tmp / "test.pptx", as_pdf=False, as_pptx=True)

    with pptx_or_pdf._export_figures() as obj:
        assert obj is not None
        plt.plot([1, 2, 3], [1, 2, 3])
        obj.add_or_export_mpl_figure(tmp / "test.png", plt.gcf())
        assert not (tmp / "test.png").exists()
        array = np.random.randint(0, 255, (1000, 1000), dtype=np.uint8)
        image = Image.fromarray(array)
        obj.add_or_export_pil_image(tmp / "test.png", image)
        assert not (tmp / "test.png").exists()
        obj.add_title("Title")
    assert (tmp / "test.pptx").exists()


def test_pdf(tmp_path):
    tmp = Path(tmp_path)
    # init PDF
    pptx_or_pdf = PptxPdf(tmp / "test.pdf", as_pdf=True)
    assert pptx_or_pdf.pdf_filename == tmp / "test.pdf"

    # create random page
    page = pptx_or_pdf._make_pdf(tmp / "test2.pdf")
    assert page is not None
    # export title
    pptx_or_pdf._add_title("Title", pdf=page)
    # export figure
    plt.plot([1, 2, 3], [1, 2, 3])
    pptx_or_pdf._add_or_export_mpl_figure(tmp / "test.png", plt.gcf(), pdf=page)
    assert not (tmp / "test.png").exists()
    # export PIL figure
    array = np.random.randint(0, 255, (1000, 1000), dtype=np.uint8)
    image = Image.fromarray(array)
    pptx_or_pdf._add_or_export_pil_image(tmp / "test.png", image, pdf=page)
    assert not (tmp / "test.png").exists()
    pptx_or_pdf._add_title("Title", pdf=page)

    pptx_or_pdf._save(pdf=page)
    assert (tmp / "test2.pdf").exists()


@pytest.mark.skipif(not HAS_PPTX, reason="pptx not installed")
def test_pptx(tmp_path):
    tmp = Path(tmp_path)
    # init PDF
    pptx_or_pdf = PptxPdf(tmp / "test.pptx", as_pptx=True)
    assert pptx_or_pdf.pptx_filename == tmp / "test.pptx"

    # create random page
    page = pptx_or_pdf._make_pptx(tmp / "test2.pptx")
    assert page is not None
    # export title
    pptx_or_pdf._add_title("Title", pptx=page)
    # export figure
    plt.plot([1, 2, 3], [1, 2, 3])
    pptx_or_pdf._add_or_export_mpl_figure(tmp / "test.png", plt.gcf(), pptx=page)
    assert not (tmp / "test.png").exists()
    plt.figure(figsize=(10, 3))
    plt.plot([1, 2, 3], [1, 2, 3])
    pptx_or_pdf._add_or_export_mpl_figure(tmp / "test.png", plt.gcf(), pptx=page)
    assert not (tmp / "test.png").exists()
    # export PIL figure
    array = np.random.randint(0, 255, (1000, 1000), dtype=np.uint8)
    image = Image.fromarray(array)
    pptx_or_pdf._add_or_export_pil_image(tmp / "test.png", image, pptx=page)
    assert not (tmp / "test.png").exists()
    pptx_or_pdf._add_title("Title", pptx=page)

    pptx_or_pdf._save(pptx=page)
    assert (tmp / "test2.pptx").exists()


def test_exporter_pdf(tmp_path):
    tmp = Path(tmp_path)
    # init PDF
    pptx_or_pdf = FigureExporter(tmp / "test.pdf", as_pdf=True)
    assert pptx_or_pdf.pdf_filename == tmp / "test.pdf"
    plt.plot([1, 2, 3], [1, 2, 3])
    plt.gcf().savefig(tmp / "test.png")
    plt.gcf().savefig(tmp / "test.jpg")
    plt.gcf().savefig(tmp / "test.jpeg")
    plt.close("all")
    pptx_or_pdf.export_existing(tmp)
    assert (tmp / "test.pdf").exists()

    pptx_or_pdf.export_existing(tmp, filename=tmp / "test2.pdf", clear=True)
    assert not (tmp / "test.png").exists()
    assert (tmp / "test2.pdf").exists()


@pytest.mark.skipif(not HAS_PPTX, reason="pptx not installed")
def test_exporter_pptx(tmp_path):
    tmp = Path(tmp_path)
    # init PDF
    pptx_or_pdf = FigureExporter(tmp / "test.pptx", as_pptx=True)
    assert pptx_or_pdf.pptx_filename == tmp / "test.pptx"
    plt.plot([1, 2, 3], [1, 2, 3])
    plt.gcf().savefig(tmp / "test.png")
    plt.gcf().savefig(tmp / "test.jpg")
    plt.gcf().savefig(tmp / "test.jpeg")
    plt.close("all")
    pptx_or_pdf.export_existing(tmp)
    assert (tmp / "test.pptx").exists()

    pptx_or_pdf.export_existing(tmp, filename=tmp / "test2.pptx", clear=True)
    assert not (tmp / "test.png").exists()
    assert (tmp / "test2.pptx").exists()



"""Garbage collection utilities."""

import gc
import time
import weakref

from loguru import logger

_uncollectable_history = []  # store snapshots for later analysis


def _gc_debug_callback(phase, info):
    # We only care after GC finishes
    if phase != "stop":
        return

    gen = info["generation"]
    n_unc = info.get("uncollectable", 0)

    if n_unc > 0:
        ts = time.time()
        print(f"[gc debug] gen {gen}: {n_unc} uncollectable objects after collection at {ts}")

        # Snapshot what's currently uncollectable
        snapshot = []
        for obj in gc.garbage:
            snapshot.append(
                {
                    "type": type(obj),
                    "repr": repr(obj)[:300],
                    "id": id(obj),
                    # store weakref so we don't keep it alive ourselves
                    "weak": weakref.ref(obj),
                }
            )

        _uncollectable_history.append(
            {
                "time": ts,
                "generation": gen,
                "snapshot": snapshot,
            }
        )

        # Optional: print a little detail to console for quick triage
        for entry in snapshot[:5]:  # don't spam if it's huge
            logger.warning("   ->", entry["type"], entry["repr"])


def install_gc_debugging() -> None:
    """Install GC debugging callback to track uncollectable objects."""
    # Register the callback (only once!)
    if _gc_debug_callback not in gc.callbacks:
        gc.callbacks.append(_gc_debug_callback)

    # Make sure gc will actually put uncollectables in gc.garbage
    gc.set_debug(gc.DEBUG_UNCOLLECTABLE)



"""Container classes."""

from __future__ import annotations

import time
import typing as ty
from abc import abstractmethod
from keyword import iskeyword
from typing import overload

from koyo.utilities import is_valid_python_name

_T = ty.TypeVar("_T")
_K = ty.TypeVar("_K")


def is_valid_python_name(name):
    """Check whether the key is valid identifier."""
    return name.isidentifier() and not iskeyword(name)


class AttributeDict:
    """Simple wrapper around schema dictionary."""

    def __init__(self, data: dict):
        self._data = data

    def __repr__(self):
        return repr(self._data)

    def __getitem__(self, key: str | int) -> ty.Any:
        """Get item id."""
        return self._data[key]

    def __getattr__(self, item: str | int) -> ty.Any:
        # allow access to group members via dot notation
        try:
            return self.__getitem__(item)
        except KeyError as err:
            raise AttributeError from err

    def __dir__(self):
        # noinspection PyUnresolvedReferences
        base = super().__dir__()
        keys = sorted(set(base + list(self._data.keys())))
        keys = [k for k in keys if is_valid_python_name(k)]
        return keys

    def _ipython_key_completions_(self):
        return sorted(self)


class MutableSequence(ty.MutableSequence[_T]):
    """List object that holds objects with optional check if it exists."""

    def __init__(self, data: ty.Iterable[_T] = ()):
        self._list: list[_T] = []
        if data:
            self.extend(data)

    def insert(self, index: int, value: _T) -> None:
        """Insert formula."""
        self._list.insert(index, self._check(value))

    @ty.overload
    def __getitem__(self, i: int) -> _T: ...

    @ty.overload
    def __getitem__(self, s: slice) -> ty.MutableSequence[_T]: ...

    def __getitem__(self, i: int) -> _T:
        return self._list[i]

    @ty.overload
    def __setitem__(self, i: int, o: _T) -> None: ...

    @ty.overload
    def __setitem__(self, s: slice, o: ty.Iterable[_T]) -> None: ...

    def __setitem__(self, key, value) -> None:
        if isinstance(key, slice):
            if not isinstance(value, ty.Iterable):
                raise TypeError("Can only assign an iterable to slice.")
            self._list[key] = [self._check(v) for v in value]
        else:
            self._list[key] = self._check(value)

    @ty.overload
    def __delitem__(self, i: int) -> None: ...

    @ty.overload
    def __delitem__(self, i: slice) -> None: ...

    def __delitem__(self, i) -> None:
        del self._list[i]

    def __len__(self) -> int:
        return len(self._list)

    def __repr__(self):
        return repr(self._list)

    def __eq__(self, other: ty.Any):
        return self._list == other

    def __hash__(self) -> int:
        # it's important to add this to allow this object to be hashable
        # given that we've also reimplemented __eq__
        return id(self)

    def _check(self, value: _T) -> _T:
        """Check whether duplicate items exist in the list."""
        return value

    def iter(self):
        """Iterator."""
        yield from self._list


class MutableMapping(ty.MutableMapping[_K, _T]):
    """Mutable mapping instance."""

    def __init__(self, data: dict[_K, _T] | None = None):
        self._dict: dict[_K, _T] = {}
        if data is not None:
            self.update(data)

    def __dir__(self) -> list[str]:
        # noinspection PyUnresolvedReferences
        base = super().__dir__()
        keys = sorted(set(base + list(self) + list(self._dict.keys())))  # type: ignore[operator]
        keys = [k for k in keys if is_valid_python_name(k)]
        return keys

    def _ipython_key_completions_(self) -> list[str]:
        return sorted(self)

    def __setitem__(self, k: _K, v: _T) -> None:
        self._dict[k] = v

    def __delitem__(self, v: _K) -> None:
        del self._dict[v]

    def __getitem__(self, k: _K) -> _T:
        return self._dict[k]

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self) -> ty.Iterator[_K]:
        yield from self._dict.keys()


class SizedDict(ty.OrderedDict):
    """Sized dictionary."""

    def __init__(self, *args: dict[_K, _T], maxsize: int = -1, **kwargs: ty.Any):
        self._maxsize = maxsize
        super().__init__(*args, **kwargs)

    def _check_size(self) -> None:
        if self._maxsize > 0:
            while len(self) > self._maxsize:
                self.popitem(last=False)

    def __setitem__(self, key: _K, value: _T) -> None:
        super().__setitem__(key, value)
        self._check_size()

    def update(self, m: dict[_K, _T], **kwargs: ty.Any) -> None:
        """Override update."""
        super().update(m, **kwargs)
        self._check_size()


class ShortLivedDict(dict):
    """Short-lived dictionary that expires after a given time-to-live (TTL)."""

    def __init__(self, ttl: float = 5.0):
        super().__init__()
        self.ttl = ttl
        self._created = time.time()

    def __getitem__(self, key: _K) -> _T:
        self._check_expired()
        return super().__getitem__(key)

    def __setitem__(self, key: _K, value: _T) -> None:
        self._check_expired()
        return super().__setitem__(key, value)

    def _expired(self) -> bool:
        return time.time() - self._created > self.ttl

    def _check_expired(self):
        if self._expired():
            super().clear()
            self._created = time.time()

    def get(self, key: _K, default: _T | None = None) -> _T | None:
        """Get item with default value."""
        self._check_expired()
        return super().get(key, default)

    def keys(self) -> ty.KeysView[_K]:
        """Get items."""
        self._check_expired()
        return super().keys()

    def items(self) -> ty.ItemsView[_K, _T]:
        """Get items."""
        self._check_expired()
        return super().items()

    def clear(self) -> None:
        """Clear the dictionary."""
        self._created = time.time()
        return super().clear()


class SizedList(ty.MutableSequence[_T]):
    """Sized list."""

    def __init__(self, *args: ty.Iterable[_T], maxsize: int = -1, **kwargs: ty.Any):
        self._maxsize = maxsize
        self._list: list[_T] = []
        super().__init__(*args, **kwargs)

    @overload
    @abstractmethod
    def __getitem__(self, index: int) -> _T: ...

    @overload
    @abstractmethod
    def __getitem__(self, index: slice) -> MutableSequence[_T]: ...

    def __getitem__(self, index: int | slice) -> _T | MutableSequence[_T]:
        return self._list[index]

    @overload
    @abstractmethod
    def __delitem__(self, index: int) -> None: ...

    @overload
    @abstractmethod
    def __delitem__(self, index: slice) -> None: ...

    def __delitem__(self, index: int | slice):
        del self._list[index]

    def __len__(self):
        return len(self._list)

    def _check_size(self) -> None:
        if self._maxsize > 0:
            while len(self) > self._maxsize:
                self.pop(0)

    def __setitem__(self, key: int, value: _T) -> None:
        super().__setitem__(key, value)
        self._check_size()

    def insert(self, index: int, value: _T):
        """Insert items."""
        self._list.insert(index, value)
        self._check_size()



"""TOML utility functions."""
import os
import typing as ty

import toml

from koyo.typing import PathLike
from koyo.utilities import _remove_duplicates_from_dict


def read_toml_data(filepath: PathLike):
    """Read TOML data and metadata.

    Parameters
    ----------
    filepath : str
        path to the TOML file

    Returns
    -------
    data : dict
        dictionary with TOML data
    """
    with open(filepath) as f_ptr:
        data = toml.load(f_ptr)
    return data


def write_toml_data(filepath: PathLike, obj: ty.Dict, check_existing: bool = False):
    """Write data to TOML file.

    Parameters
    ----------
    filepath : PathLike
        path to TOML file
    obj : dict
        object containing data
    check_existing : bool, optional
        if True, existing JSON file data will be merged with the new data object
    """
    if not check_existing or not os.path.exists(filepath):
        with open(filepath, "w") as f_ptr:
            toml.dump(obj, f_ptr)
    else:
        with open(filepath, "r+") as f_ptr:
            try:
                data = toml.load(f_ptr)
            except toml.decoder.TomlDecodeError:
                data = {}

            if isinstance(data, list):
                data.extend(obj)
            elif isinstance(data, dict):
                data.update(obj)

            # remove duplicates
            data = _remove_duplicates_from_dict(data)

            f_ptr.seek(0)  # rewind
            toml.dump(data, f_ptr)
            f_ptr.truncate()



import time

import pytest
from koyo.timer import format_time, time_average, time_loop


@pytest.mark.parametrize(
    "value, expected",
    [
        (0.01, "us"),
        (0.1, "ms"),
        (0.5, "s"),
        (1, "s"),
        (60, "s"),
        (75, "min"),
        (3654, "hr"),
    ],
)
def test_format_time(value, expected):
    """Test 'format_time'"""
    result = format_time(value)
    assert expected in result


def test_time_average():
    """Test 'time_average'"""
    t_start = time.time()
    time.sleep(0.01)
    result = time_average(t_start, 1)
    assert "Avg:" in result
    assert "Tot:" in result
    assert result.startswith("[")
    assert result.endswith("]")


def test_time_loop():
    """Test 'time_loop'"""
    t_start = time.time()
    time.sleep(0.01)
    result = time_loop(t_start, 0, 0)
    assert "Avg:" in result
    assert "Rem:" in result
    assert "Tot:" in result
    assert "%" in result

    result = time_loop(t_start, 0, 0, as_percentage=False)
    assert "%" not in result
    assert result.startswith("[")
    assert result.endswith("]")



import os

import pytest
import toml
from koyo.toml import read_toml_data, write_toml_data


def test_read_01(tmp_path):
    path = os.path.join(tmp_path, "config.toml")
    test_dict = {"version": 0.1, "data": [1, 2, 3]}
    with open(path, "w") as f_ptr:
        toml.dump(test_dict, f_ptr)

    results = read_toml_data(path)
    assert test_dict == results


def test_read_02(tmp_path):
    path = os.path.join(tmp_path, "config.toml")

    with pytest.raises(OSError):
        _ = read_toml_data(path)


def test_write_write_new(tmp_path):
    path = os.path.join(tmp_path, "config.toml")

    test_dict = {"version": 0.1, "data": [4, 5, 6]}
    write_toml_data(path, test_dict)

    # load data
    with open(path) as f_ptr:
        results = toml.load(f_ptr)
    assert test_dict == results

    # change data - should not be overriden
    test_dict = {"data": [4, 5, 6]}
    write_toml_data(path, test_dict, check_existing=False)

    # load data
    with open(path) as f_ptr:
        results = toml.load(f_ptr)
    assert test_dict == results


def test_write_update_existing(tmp_path):
    path = os.path.join(tmp_path, "config.toml")

    test_dict = {"version": 0.1, "data": [4, 5, 6]}

    write_toml_data(path, test_dict)
    with open(path) as f_ptr:
        results = toml.load(f_ptr)
    assert test_dict == results

    add_data_dict = {"newkey": 42}
    write_toml_data(path, add_data_dict, check_existing=True)
    with open(path) as f_ptr:
        results = toml.load(f_ptr)
    test_dict.update(add_data_dict)
    assert add_data_dict != results
    assert test_dict == results



"""Path utilities."""

import os
import shutil
import typing as ty
from pathlib import Path

from loguru import logger

from koyo.typing import PathLike


def copy_file(src_path: str, dst_path: str) -> None:
    """Copy a file from src_path to dst_path using zero-copy os.sendfile()."""
    # Try zero-copy if available
    if hasattr(os, "sendfile"):
        with open(src_path, "rb") as fsrc, open(dst_path, "wb") as fdst:
            size = os.fstat(fsrc.fileno()).st_size
            offset = 0
            while offset < size:
                # Send up to 64 MB per call to avoid platform-specific limits
                sent = os.sendfile(fdst.fileno(), fsrc.fileno(), offset, min(size - offset, 64 * 1024 * 1024))
                if sent == 0:
                    break
                offset += sent
        return
    # Fallback to shutil.copyfile if zero-copy is not available
    try:
        shutil.copyfile(src_path, dst_path)
    except FileNotFoundError:
        logger.error(f"Source file '{src_path}' does not exist.")


def mglob(path: Path, *patterns: str, recursive: bool = False) -> ty.Generator[Path, None, None]:
    """Glob multiple patterns."""
    if not path.exists():
        return
    for pattern in patterns:
        if recursive:
            yield from path.rglob(pattern)
        else:
            yield from path.glob(pattern)


def dir_iter(path: PathLike) -> ty.Generator[Path, None, None]:
    """Iterate over directory."""
    path = Path(path)
    for item in path.iterdir():
        if item.is_dir():
            yield item


def uri_to_path(uri: str) -> Path:
    """Convert URI to path."""
    try:
        from urllib.parse import unquote, urlparse
        from urllib.request import url2pathname
    except ImportError:
        # backwards compatability
        from urllib import unquote, url2pathname

        from urlparse import urlparse

    parsed = urlparse(uri)
    host = f"{os.path.sep}{os.path.sep}{parsed.netloc}{os.path.sep}"
    return Path(os.path.normpath(os.path.join(host, url2pathname(unquote(parsed.path)))))


def get_copy_path(path: PathLike) -> Path:
    """Get copy path."""
    path = Path(path)
    if not path.exists():
        return path
    i = 1
    while True:
        new_path = path.parent / f"{path.stem}_copy{i}{path.suffix}"
        if not new_path.exists():
            break
        i += 1
    return new_path


def remove_file(file: PathLike) -> None:
    """Remove file."""
    file = Path(file)
    if file.exists():
        if file.is_file():
            try:
                file.unlink()
                logger.trace(f"Removed '{file}'")
            except Exception as e:
                logger.error(f"Failed to remove '{file}'. Reason: {e}")
        else:
            logger.error(f"Failed to remove '{file}'. Reason: Not a file - it's a directory.")


def empty_directory(path: PathLike) -> None:
    """Recursively clear directory."""
    path = Path(path)
    if not path.exists():
        return
    try:
        for filename in os.listdir(path):
            file_path = os.path.join(path, filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                    logger.trace(f"Deleted '{file_path}'")
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
                    logger.trace(f"Deleted '{file_path}'")
            except Exception as e:
                logger.warning(f"Failed to delete {file_path}. Reason: {e}")
    except Exception as e:
        logger.error(f"Failed to empty directory '{path}'. Reason: {e}")
    try:
        shutil.rmtree(path, ignore_errors=True)
        logger.trace(f"Deleted '{path}'")
    except Exception as e:
        logger.error(f"Failed to delete '{path}'. Reason: {e}")


remove_directory = empty_directory


def move_directory(src: PathLike, dest: PathLike) -> None:
    """Move directory."""
    src = Path(src)
    dest = Path(dest)

    if src == dest:
        return

    try:
        shutil.move(src, dest)
        logger.trace(f"Moved '{src}' to '{dest}'")
    except Exception as e:
        logger.error(f"Failed to move '{src}' to '{dest}'. Reason: {e}")


def open_directory(path: PathLike) -> None:
    """Open directory."""
    import webbrowser

    if not os.path.isdir(path):
        path = os.path.dirname(path)
    webbrowser.open(str(path))


def open_directory_alt(path: PathLike, *which: str) -> None:
    """Open directory."""
    import platform
    import subprocess

    if str(path).startswith("file://"):
        path = str(path)[7:]
        path = path.lstrip("/")
    path = Path(path)
    if which:
        path = path.joinpath(*which)

    if platform.system() == "Windows":
        if path.exists():
            if path.is_file():
                subprocess.call(["explorer", "/select,", str(path)])
            else:
                subprocess.call(["explorer", str(path)])
        else:
            subprocess.call(["explorer", os.path.dirname(path)])
    elif platform.system() == "Darwin":
        if path.is_file():
            subprocess.call(["open", "-R", str(path)])
        else:
            subprocess.Popen(["open", str(path)])
    else:
        subprocess.Popen(["xdg-open", str(path)])


def create_directory(*path: str) -> Path:
    """Create directory.

    Parameters
    ----------
    path : str
        directory path
    """
    path = Path(os.path.join(*path))
    if not path.exists():
        try:
            path.mkdir(parents=True)
        except OSError:
            logger.warning("Failed to create directory")
        finally:
            return path


def is_network_path(path: PathLike) -> bool:
    """Check if the path is a network path."""
    if os.name == "nt":
        return is_network_path_win(path)
    return is_network_path_unix(path)


def is_network_path_win(path: PathLike) -> bool:
    """Check if the path is a network path."""
    import os

    import win32file

    # Get the root path (e.g., "C:\\" from "C:\\path\\to\\file")
    root_path = os.path.splitdrive(path)[0] + "\\"

    # Get the drive type
    drive_type = win32file.GetDriveType(root_path)

    # Check if the drive type is network
    return bool(drive_type == win32file.DRIVE_REMOTE)


def get_mount_point(path: PathLike) -> str:
    """Finds the mount point for a given path."""
    path = os.path.abspath(path)
    while not os.path.ismount(path):
        path = os.path.dirname(path)
    return path


def is_network_path_unix(path: PathLike) -> bool:
    """Checks if the path is on a network-mounted filesystem."""
    import subprocess

    mount_point = get_mount_point(path)
    # On Linux, you might directly parse /proc/mounts for more detailed info
    # For simplicity, we use df
    df_output = subprocess.check_output(["df", mount_point], universal_newlines=True)
    filesystem_type = df_output.split("\n")[1].split()[0]

    # This is a simple heuristic, you may need to adjust the types based on your needs
    network_filesystems = ["nfs", "smbfs", "cifs", "fuse.sshfs"]

    return any(fs_type in filesystem_type for fs_type in network_filesystems)



"""Mixin class for figure exports."""

from __future__ import annotations

import typing as ty
from contextlib import contextmanager
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from loguru import logger
from tqdm import tqdm

from koyo.pdf_mixin import PDFMixin
from koyo.pptx_mixin import PPTXMixin
from koyo.system import IS_WIN
from koyo.typing import PathLike

if ty.TYPE_CHECKING:
    from matplotlib.backends.backend_pdf import PdfPages
    from PIL import Image

    try:
        from pptx.presentation import Presentation
    except ImportError:
        Presentation = None  # type: ignore[assignment,misc]


class FigureMixin(PDFMixin, PPTXMixin):
    """Mixin class for figure exports."""

    @property
    def as_pptx_or_pdf(self) -> bool:
        """Check whether export is enabled."""
        return self.as_pptx or self.as_pdf

    def _check_export(self, as_pptx: bool | None = None, as_pdf: bool | None = None) -> None:
        """Check whether pdf/pptx export is enabled."""
        if as_pptx is not None:
            self.as_pptx = as_pptx
        if as_pdf is not None:
            self.as_pdf = as_pdf
        if self.as_pptx and self.as_pdf:
            raise ValueError("Cannot export to both PDF and PPTX")

    def get_pptx_or_pdf_filename(self, filename: PathLike) -> str:
        """Get export filename."""
        filename = str(filename)
        parts = filename.rsplit(".", maxsplit=1)
        # if there are multiple ., then let's combine it using . but exclude the last one
        if len(parts) > 1:
            filename = ".".join(parts[:-1])
        if self.as_pptx:
            return f"{filename}.pptx"
        elif self.as_pdf:
            return f"{filename}.pdf"
        return filename

    def has_slides(self, pptx_filename: Path) -> bool:
        """Check whether PPTX has slides."""
        from pptx import Presentation

        if pptx_filename.exists():
            pptx = Presentation(pptx_filename)
            has_slides = bool(pptx.slides)
            return has_slides
        return False

    def get_actual_output_filename(self, output_dir: PathLike, pptx_or_pdf_path: PathLike) -> Path:
        """Get output filename, depending on the export format."""
        if self.as_pptx_or_pdf:
            return Path(pptx_or_pdf_path)
        return Path(output_dir)

    def make_directory_if_not_exporting(self, directory: PathLike) -> Path:
        """Make directory if it's not being exported to PDF/PPTX."""
        directory = Path(directory)
        if not self.as_pptx_or_pdf:
            directory.mkdir(parents=True, exist_ok=True)
        return directory

    @contextmanager
    def _export_figures(self, filename: PathLike | None = None) -> ty.Iterator[PptxPdfWrapper]:
        """Export figures."""
        if filename:
            logger.trace(f"Exporting figures to {filename}")
        if self.as_pptx:
            with self._export_pptx_figures(filename) as pptx:
                yield PptxPdfWrapper(pptx, as_pptx=self.as_pptx, as_pdf=self.as_pdf)
        elif self.as_pdf:
            with self._export_pdf_figures(filename) as pdf:
                yield PptxPdfWrapper(pdf, as_pptx=self.as_pptx, as_pdf=self.as_pdf)
        else:
            yield PptxPdfWrapper(None, as_pptx=self.as_pptx, as_pdf=self.as_pdf)

    def _add_title(self, title: str, pdf: PdfPages | None = None, pptx: Presentation | None = None) -> None:
        """Add title to pptx."""
        if self.as_pptx:
            self._add_title_to_pptx(title, pptx=pptx)
        elif self.as_pdf:
            self._add_title_to_pdf(title, pdf=pdf)

    def _add_content(
        self, content: str, title: str, pdf: PdfPages | None = None, pptx: Presentation | None = None
    ) -> None:
        """Add title to pptx."""
        if self.as_pptx:
            self._add_content_to_pptx(content, title, pptx=pptx)
        elif self.as_pdf:
            self._add_content_to_pdf(content, title, pdf=pdf)

    def _save(self, pptx: Presentation | None = None, pdf: PdfPages | None = None) -> None:
        """Save pptx/pdf."""
        if self.as_pptx:
            self._save_pptx(pptx)
        elif self.as_pdf:
            self._save_pdf(pdf)

    def _add_or_export_mpl_figure(
        self,
        filename: Path,
        fig: plt.Figure | None,
        face_color: str | np.ndarray | None = None,
        bbox_inches: str | None = "tight",
        dpi: int = 150,
        override: bool = False,
        pdf: PdfPages | None = None,
        pptx: Presentation | None = None,
        close: bool = False,
        **kwargs: ty.Any,
    ) -> None:
        """Export figure to file."""
        if fig is None:
            logger.warning("Figure is None, skipping export")
            return

        if self.as_pdf:
            self._add_mpl_figure_to_pdf(filename, fig, pdf=pdf, **kwargs)
        elif self.as_pptx:
            self._add_mpl_figure_to_pptx(filename, fig, pptx=pptx, **kwargs)
        elif override or not filename.exists():
            fig.savefig(filename, dpi=dpi, facecolor=face_color, bbox_inches=bbox_inches, **kwargs)
            if close:
                plt.close(fig)

    def _add_or_export_pil_image(
        self,
        filename: Path,
        image: Image,
        dpi: int = 150,
        fmt: str = "JPEG",
        override: bool = False,
        pdf: PdfPages | None = None,
        pptx: Presentation | None = None,
        close: bool = False,
        **kwargs: ty.Any,
    ) -> None:
        """Export PIL image to file."""
        if self.as_pdf:
            self._add_pil_image_to_pdf(filename, image, pdf=pdf, **kwargs)
        elif self.as_pptx:
            self._add_pil_image_to_pptx(filename, image, pptx=pptx, **kwargs)
        elif override or not filename.exists():
            image.save(filename, dpi=(dpi, dpi), format=fmt, **kwargs)
            if close:
                image.close()


class FigureExporter(FigureMixin):
    """Wrapper around the mixin class to provide the interface without having to subclass it."""

    def __init__(self, filename: PathLike | None = None, as_pptx: bool = False, as_pdf: bool = False) -> None:
        self.filename = Path(filename) if filename else None
        self._check_export(as_pptx=as_pptx, as_pdf=as_pdf)

    @property
    def pdf_filename(self) -> Path:
        """Return the PDF filename."""
        if not self.filename:
            raise ValueError("Filename not set")
        return Path(self.get_pptx_or_pdf_filename(self.filename))

    @property
    def pptx_filename(self) -> Path:
        """Return the PPTX filename."""
        if not self.filename:
            raise ValueError("Filename not set")
        return Path(self.get_pptx_or_pdf_filename(self.filename))

    def export_existing(
        self,
        input_dir: PathLike,
        extensions: tuple[str, ...] = ("*.png", "*.jpg", "*.jpeg"),
        filename: PathLike | None = None,
        clear: bool = False,
        silent: bool = False,
    ) -> None:
        """Export existing figures from within a directory (this includes any PNG, JPEG) files read by PIL."""
        from PIL import Image

        extensions = tuple(ext if ext.startswith("*") else f"*{ext}" for ext in extensions)
        assert any(ext.startswith("*") for ext in extensions), "Extensions must start with '*'"
        assert "*.pptx" not in extensions, "Cannot export PPTX files"
        assert "*.pdf" not in extensions, "Cannot export PDF files"

        n = 0
        with self._export_figures(filename) as pptx_or_pdf:
            for ext in extensions:
                for file in tqdm(
                    Path(input_dir).rglob(ext), desc=f"Exporting images ({ext})", disable=silent, leave=False
                ):
                    pptx_or_pdf.add_or_export_pil_image(file, Image.open(file), close=True)
                    n += 1
        # if clear is enabled, remove directory if it is empty
        if clear:
            for ext in extensions:
                for file in Path(input_dir).rglob(ext):
                    file.unlink()
            if not list(Path(input_dir).rglob("*")):
                Path(input_dir).rmdir()
            logger.trace(f"Removed empty directory '{input_dir}'")
        logger.info(f"Exported {n} images from '{input_dir}' to '{self.filename}'")


class PptxPdfWrapper:
    """Wrapper class that handles both PPTX and PDF exports."""

    def __init__(
        self, ppt_or_pdf: PdfPages | Presentation | None = None, as_pptx: bool = False, as_pdf: bool = False
    ) -> None:
        self.ppt_or_pdf = ppt_or_pdf
        self.as_pptx = as_pptx
        self.as_pdf = as_pdf

    def make_directory_if_not_exporting(self, directory: PathLike) -> Path:
        """Make directory if it's not being exported to PDF/PPTX."""
        directory = Path(directory)
        if not self.as_pptx_or_pdf:
            directory.mkdir(parents=True, exist_ok=True)
        return directory

    def figure_exists(self, filename: Path, override: bool = False) -> bool:
        """Check whether figure exists."""
        return (filename.exists() and not override) and not self.as_pptx_or_pdf

    @property
    def as_pptx_or_pdf(self) -> bool:
        """Check whether export is enabled."""
        return self.as_pptx or self.as_pdf

    def add_title(self, title: str) -> None:
        """Add title slide or page."""
        if self.as_pdf:
            from koyo.pdf_mixin import add_title_to_pdf

            add_title_to_pdf(self.ppt_or_pdf, title)
        elif self.as_pptx:
            from koyo.pptx_mixin import add_title_to_pptx

            add_title_to_pptx(self.ppt_or_pdf, title)  # type: ignore[arg-type]

    def add_content(self, content: str, title: str = "") -> None:
        """Add title slide or page."""
        if self.as_pptx:
            from koyo.pptx_mixin import add_content_to_pptx

            add_content_to_pptx(self.ppt_or_pdf, content, title)  # type: ignore[arg-type]
        elif self.as_pdf:
            from koyo.pdf_mixin import add_content_to_pdf

            add_content_to_pdf(self.ppt_or_pdf, content, title)

    def add_or_export_pil_image(
        self,
        filename: PathLike,
        image: Image,
        dpi: int = 150,
        fmt: str = "JPEG",
        override: bool = False,
        close: bool = False,
        title: str = "",
        **kwargs: ty.Any,
    ) -> None:
        """Add or export PIL image."""
        if self.as_pdf:
            from koyo.pdf_mixin import add_pil_image_to_pdf

            add_pil_image_to_pdf(self.ppt_or_pdf, filename, image, override=override, **kwargs)
        elif self.as_pptx:
            from koyo.pptx_mixin import add_pil_image_to_pptx

            add_pil_image_to_pptx(self.ppt_or_pdf, filename, image, override=override, title=title, **kwargs)
        elif override or not Path(filename).exists():
            image.save(filename, dpi=(dpi, dpi), format=fmt, **kwargs)
            if close:
                image.close()

    def add_or_export_mpl_figure(
        self,
        filename: PathLike,
        fig: plt.Figure | None,
        face_color: str | np.ndarray | None = None,
        bbox_inches: str | None = "tight",
        dpi: int = 150,
        override: bool = False,
        close: bool = False,
        title: str = "",
        **kwargs: ty.Any,
    ) -> None:
        """Add or export matplotlib figure."""
        if fig is None:
            logger.warning("Figure is None, skipping export")
            return
        if self.as_pdf:
            from koyo.pdf_mixin import add_mpl_figure_to_pdf

            add_mpl_figure_to_pdf(
                self.ppt_or_pdf,
                filename,
                fig,
                face_color=face_color,
                bbox_inches=bbox_inches,
                dpi=dpi,
                override=override,
                close=close,
                **kwargs,
            )
        elif self.as_pptx:
            from koyo.pptx_mixin import add_mpl_figure_to_pptx

            add_mpl_figure_to_pptx(
                self.ppt_or_pdf,
                filename,
                fig,
                face_color=face_color,
                bbox_inches=bbox_inches,
                dpi=dpi,
                override=override,
                close=close,
                title=title,
                **kwargs,
            )
        elif override or not Path(filename).exists():
            face_color = face_color if face_color is not None else fig.get_facecolor()
            filename = _ensure_filename_is_not_too_long(filename)
            fig.savefig(filename, dpi=dpi, facecolor=face_color, bbox_inches=bbox_inches, **kwargs)
            if close:
                plt.close(fig)


def _ensure_filename_is_not_too_long(filenme: PathLike) -> Path:
    """Ensures on Windows that filename is not too long."""
    if not IS_WIN:
        return Path(filenme)
    filename = Path(filenme)
    n = len(str(filename))
    if n > 250:
        parent = filename.parent
        suffix = filename.suffix
        if n - len(str(parent)) > 250:
            raise ValueError("Filename is too long")
        max_length = 250 - len(str(parent)) - len(suffix)
        if max_length > len(filename.stem):
            max_length = len(filename.stem)
        filename_ = filename.stem[0:max_length] + suffix
        logger.trace(f"Filename is too long, truncating to {filename_} from {filename.name}")
        return parent / filename_
    return filename



"""Module with color functions."""

from __future__ import annotations

import colorsys
import random
import typing as ty
import warnings
from ast import literal_eval

import numpy as np

if ty.TYPE_CHECKING:
    from matplotlib.colors import Colormap as MplColormap


# All parsable input color types that a user can provide
ColorType = ty.Union[list, tuple, np.ndarray, str]


def mpl_to_rgba_255(color: str | tuple | list) -> tuple[int, int, int, int]:
    """Convert color from matplotlib to RGBA 255-scale."""
    from matplotlib.colors import to_rgba_array

    return tuple(to_rgba_array(color)[0].astype(np.uint8) * 255)


def rgb_1_to_hex(color: tuple | list):
    """Convert RGB to Hex."""
    color = list(color)[0:3]
    return "#" + "".join([f"{int(c * 255):02x}" for c in color])


def rgb_255_to_1(color: tuple | list, decimals: int = 3):
    """Convert color that is in RGB (255-scale) to RGB (1-scale).

    Parameters
    ----------
    color : list or np.array
        color
    decimals : int, optional
        number of decimal spaces

    Returns
    -------
    rgbList : list
        color in 1-scale
    """
    # make sure color is an rgb format and not string format
    try:
        color = literal_eval(color)
    except Exception:
        color = color

    color = np.array(color)
    if color.shape[0] not in [3, 4]:
        raise ValueError("Color must have shape (3,) or (4,)")
    if color.max() > 255:
        raise ValueError("Color or transparency cannot be larger than 255")
    if color.min() < 0:
        raise ValueError("Color or transparency cannots be larger than 0")

    color = (color / 255).round(decimals)

    return list(color)


def hex_to_rgb_1(hex_str, decimals: int = 3, with_alpha: bool = False):
    """Convert hex color to rgb color in 1-scale."""
    hex_color = hex_str.lstrip("#")
    hlen = len(hex_color)
    rgb = tuple(int(hex_color[i : i + int(hlen / 3)], 16) for i in range(0, int(hlen), int(hlen / 3)))
    rgb = [np.round(rgb[0] / 255.0, decimals), np.round(rgb[1] / 255.0, decimals), np.round(rgb[2] / 255.0, decimals)]
    if with_alpha:
        return [*rgb, 1.0]
    return rgb


def hex_to_rgb_255(hex_str):
    """Convert hex color to rgb color in 255-scale."""
    hex_color = hex_str.lstrip("#")
    hlen = len(hex_color)
    rgb = [int(hex_color[i : i + int(hlen / 3)], 16) for i in range(0, int(hlen), int(hlen / 3))]
    return rgb


def rgb_to_hex(colors, multiplier: int = 255) -> str:
    """Convert list/tuple of colors to hex."""
    return f"#{int(colors[0] * multiplier):02x}{int(colors[1] * multiplier):02x}{int(colors[2] * multiplier):02x}"


def hex_to_rgb(hex_str, decimals=3, alpha: float | None = None):
    """Convert hex color to numpy array."""
    hex_color = hex_str.lstrip("#")
    hex_len = len(hex_color)
    rgb = [int(hex_color[i : i + int(hex_len / 3)], 16) for i in range(0, int(hex_len), int(hex_len / 3))]
    if alpha is not None:
        if alpha == 1:
            warnings.warn(
                "The provided alpha value is equal to 1 - this function accepts values in 0-255 range.", stacklevel=2
            )
        rgb.append(alpha)
    return np.round(np.asarray(rgb) / 255, decimals)


def get_random_hex_color() -> str:
    """Return random hex color."""
    return f"#{random.randint(0, 0xFFFFFF):06x}"


def get_next_color(n: int, other_colors: list[str] | None = None) -> str:
    """
    Get the next color based on the index, avoiding duplicates in `other_colors`.

    Parameters
    ----------
        n (int): The index of the desired color.
        other_colors (list[str] | None): A list of colors to avoid, default is None.

    Returns
    -------
        str: The next distinct color in lowercase hex format.
    """
    if other_colors is None:
        other_colors = []

    # Predefined list of colors
    colors = [
        "#ff0000",
        "#00ff00",
        "#0000ff",
        "#ffff00",
        "#ff00ff",
        "#00ffff",
        "#ff4500",
        "#ff69b4",
        "#1e90ff",
        "#32cd32",
        "#ffd700",
        "#40e0d0",
        "#ff6347",
        "#7b68ee",
        "#adff2f",
        "#87ceeb",
    ]

    # Extend colors if `n` exceeds predefined list
    if n >= len(colors):
        additional_count = n - len(colors) + 50
        hue_step = 1 / (additional_count + 1)
        for i in range(additional_count):
            h = (i + 1) * hue_step
            s, l = 0.7, 0.5
            r, g, b = colorsys.hls_to_rgb(h, l, s)
            colors.append(f"#{int(r * 255):02x}{int(g * 255):02x}{int(b * 255):02x}")

    # Get the nth color and check for duplicates
    color = colors[n].lower()
    if color in (c.lower() for c in other_colors):
        return get_next_color(n + 1, other_colors=other_colors)
    return color


def generate_distinct_colors(starting_colors: list[str], n_colors: int) -> list[str]:
    """
    Generate a list of n_colors distinct colors starting from the given list.

    Parameters
    ----------
        starting_colors (List[str]): List of predefined colors in hex format.
        n_colors (int): The total number of distinct colors required.

    Returns
    -------
        List[str]: List of n_colors in hex format.
    """

    def hsl_to_hex(h, s, l):
        r, g, b = colorsys.hls_to_rgb(h, l, s)
        return f"#{int(r * 255):02x}{int(g * 255):02x}{int(b * 255):02x}"

    # Convert predefined colors to HSL and collect them
    starting_hsl = []
    for hex_color in starting_colors:
        hex_color = hex_color.lstrip("#")
        r, g, b = (int(hex_color[i : i + 2], 16) / 255.0 for i in (0, 2, 4))
        h, l, s = colorsys.rgb_to_hls(r, g, b)
        starting_hsl.append((h, s, l))

    # Generate new colors if needed
    additional_colors_needed = max(0, n_colors - len(starting_colors))
    if additional_colors_needed > 0:
        step = 1 / (additional_colors_needed + 1)
        for i in range(additional_colors_needed):
            h = (i + 1) * step  # Spread hues evenly
            s, l = 0.7, 0.5  # Fixed saturation and lightness for distinctiveness
            starting_hsl.append((h, s, l))

    # Convert all HSL colors back to hex
    result_colors = starting_colors[:]
    result_colors += [hsl_to_hex(h, s, l) for h, s, l in starting_hsl[len(starting_colors) :]]
    return result_colors[:n_colors]


def colormap_to_hex(colormap: MplColormap) -> list[str]:
    """Convert mpl colormap to hex."""
    return [rgb_to_hex(colormap(i)) for i in range(colormap.N)]


def get_colors_from_colormap(colormap: str, n_colors: int, is_reversed: bool = False) -> list[str]:
    """Get list of colors from colormap."""
    import matplotlib.cm

    if is_reversed and not colormap.endswith("_r"):
        colormap += "_r"

    return colormap_to_hex(matplotlib.cm.get_cmap(colormap, n_colors))


def find_text_color(base_color, dark_color="black", light_color="white", coef_choice=0):
    """
    Takes a background color and returns the appropriate light or dark text color.
    Users can specify the dark and light text color, or accept the defaults of 'black' and 'white'
    base_color: The color of the background. This must be
        specified in RGBA with values between 0 and 1 (note, this is the default
        return value format of a call to base_color = cmap(number) to get the
        color corresponding to a desired number). Note, the value of `A` in RGBA
        is not considered in determining light/dark.
    dark_color: Any valid matplotlib color value.
        Function will return this value if the text should be colored dark
    light_color: Any valid matplotlib color value.
        Function will return this value if thet text should be colored light.
    coef_choice: slightly different approaches to calculating brightness. Currently two options in
        a list, user can enter 0 or 1 as list index. 0 is default.
    """
    # Coefficients:
    # option 0: http://www.nbdtech.com/Blog/archive/2008/04/27/Calculating-the-Perceived-Brightness-of-a-Color.aspx
    # option 1: http://stackoverflow.com/questions/596216/formula-to-determine-brightness-of-rgb-color
    coef_options = [
        np.array((0.241, 0.691, 0.068, 0)),
        np.array((0.299, 0.587, 0.114, 0)),
    ]

    coefs = coef_options[coef_choice]
    rgb = np.array(base_color) * 255
    brightness = np.sqrt(np.dot(coefs, rgb**2))

    # Threshold from option 0 link; determined by trial and error.
    # base is light
    if brightness > 130:
        return dark_color
    return light_color


def make_listed_colormap(colors: list[str], is_vispy: bool = False) -> MplColormap:
    """Make listed colormap."""
    from matplotlib.colors import LinearSegmentedColormap
    from vispy.color.colormap import Colormap

    if is_vispy:
        colors.insert(0, "#FFFFFF")

    colormap = LinearSegmentedColormap.from_list("colormap", colors, len(colors))
    if is_vispy:
        mpl_colors = colormap(np.linspace(0, 1, len(colors)))
        mpl_colors[0][-1] = 0  # first color is white with alpha=0
        return Colormap(mpl_colors)
    return colormap


def _check_color_dim(val):
    """Ensures input is Nx4.

    Parameters
    ----------
    val : np.ndarray
        A color array of possibly less than 4 columns

    Returns
    -------
    val : np.ndarray
        A four columns version of the input array. If the original array
        was a missing the fourth channel, it's added as 1.0 values.
    """
    val = np.atleast_2d(val)
    if val.shape[1] not in (3, 4):
        strval = str(val)
        if len(strval) > 100:
            strval = strval[:97] + "..."
        raise RuntimeError(f"Value must have second dimension of size 3 or 4. Got `{strval}`, shape={val.shape}")

    if val.shape[1] == 3:
        val = np.column_stack([val, np.float32(1.0)])
    return val


def rgbs_to_hex(rgbs: ty.Sequence) -> np.ndarray:
    """Convert RGB to hex quadruplet.

    Taken from vispy with slight modifications.

    Parameters
    ----------
    rgbs : Sequence
        A list-like container of colors in RGBA format with values
        between [0, 1]

    Returns
    -------
    arr : np.ndarray
        An array of the hex representation of the input colors

    """
    rgbs = _check_color_dim(rgbs)
    return np.array(
        [f"#{'%02x' * 4}" % tuple((255 * rgb).astype(np.uint8)) for rgb in rgbs],
        "|U9",
    )


def transform_color(color: ColorType) -> np.ndarray:
    """Transform color."""
    if isinstance(color, str):
        if color.startswith("#"):
            return np.atleast_2d(hex_to_rgb_1(color, with_alpha=True))
        if color.startswith("0x"):
            return np.atleast_2d(hex_to_rgb_1(color[2:], with_alpha=True))
        if color.startswith("rgb"):
            return np.atleast_2d(literal_eval(color))
        if color.startswith("rgba"):
            return np.atleast_2d(literal_eval(color))
        raise ValueError(f"Invalid color string: {color}")
    if isinstance(color, (list, tuple)):
        color = np.atleast_2d(color)
        if color.max() > 1:
            color = color / 255.0
        if color.shape[1] == 4:
            return color
        if color.shape[1] == 3:
            return np.concatenate((color, np.ones((color.shape[0], 1))), axis=1)
        raise ValueError(f"Invalid color list/tuple: {color}")
    if isinstance(color, np.ndarray):
        if color.ndim == 1:
            if color.shape[0] == 3:
                return np.concatenate((color, [1]))
            if color.shape[0] == 4:
                return color
        raise ValueError(f"Invalid color array: {color}")
